/**
 * Ono Project
 *
 * File:         DataTransferRunner.java
 * RCS:          $Id: OnoDataTransferManager.java,v 1.15 2007/09/21 16:58:02 drc915 Exp $
 * Description:  DataTransferRunner class (see below)
 * Author:       David Choffnes
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Aug 23, 2006 at 1:15:00 PM
 * Language:     Java
 * Package:      edu.northwestern.ono.experiment
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2006, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.ono.net;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.OutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.PrintStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Random;
import java.util.Set;
import java.util.TimeZone;
import java.util.TreeMap;
import java.util.Vector;
import java.util.Map.Entry;

import org.gudy.azureus2.core3.util.Constants;
import org.xbill.DNS.Address;

import com.Ostermiller.util.CircularByteBuffer;

import edu.northwestern.ono.MainGeneric;
import edu.northwestern.ono.MainPlanetLab;
import edu.northwestern.ono.api.TransferManager;
import edu.northwestern.ono.connection.IConnectionHandler;
import edu.northwestern.ono.connection.IConnectionListener;
import edu.northwestern.ono.connection.IMessageException;
import edu.northwestern.ono.connection.IOnoConnection;
import edu.northwestern.ono.connection.IOnoConnectionManager;
import edu.northwestern.ono.dht.IDDBReadAction;
import edu.northwestern.ono.dht.IDistributedDatabase;
import edu.northwestern.ono.dht.IDistributedDatabaseEvent;
import edu.northwestern.ono.dns.Digger;
import edu.northwestern.ono.experiment.DataTransferRunner;
import edu.northwestern.ono.experiment.TraceRouteRunner;
import edu.northwestern.ono.experiment.DataTransferRunner.EdgeServerListHandler;
import edu.northwestern.ono.position.CDNClusterFinder;
import edu.northwestern.ono.position.OnoPeerManager.OnoPeer;
import edu.northwestern.ono.stats.Statistics;
import edu.northwestern.ono.timer.ITimer;
import edu.northwestern.ono.timer.ITimerEvent;
import edu.northwestern.ono.timer.ITimerEventPerformer;
import edu.northwestern.ono.util.Pair;
import edu.northwestern.ono.util.PluginInterface;
import edu.northwestern.ono.util.Util;


/**
 * @author David Choffnes &lt;drchoffnes@cs.northwestern.edu&gt;
 *
 * The DataTransferRunner class uses detouring during a data
 * transfer experiment.
 */
public class OnoDataTransferManager implements TransferManager, IConnectionListener,
     IDDBReadAction, IConnectionHandler {
    private static OnoDataTransferManager self;
    private static PluginInterface pi;
    public static final boolean DEBUG = false;

    /** message type constant for requesting a list of edges */
    public static final byte TYPE_EDGES = 1;

    /** message type for data sent directly between source and dest */
    public static final byte TYPE_PRIMARY_PATH = 2;

    /** message type for race between direct and one-hop path */
    public static final byte TYPE_RACE = 3;

    /** message type for receiver sending race result */
    static public final byte TYPE_RACE_RESULT = 4;

    /** message type for list of edges */
    public static final byte TYPE_EDGE_LIST = 5;

    /** message type for requesting peer list from dest */
    public static final byte TYPE_PEER_REQUEST = 6;

    /** message type for sending peer list to source */
    public static final byte TYPE_PEER_LIST = 7;

    // TODO send acks along paths their data came from...
    /** message type for message-receipt acknowledgement */
    public static final byte TYPE_ACK = 8;
    
    /** message type sent to dest to announce an incoming experiment */
    public static final byte TYPE_EXP_ANNOUNCE = 9;

    /** "ip" value for direct path between two hosts */
    private static final byte[] DIRECT_BYTES = { 0, 0, 0, 0 };

    /** size of data for race */
    private static final long SIZE_EVAL = 3 * 1000 * 1000;

    /** size of non-warmup data packet*/
    private static final int NON_WARMPUP_SIZE = 64000;

    /** size of warmup packet */
    public static final int WARMUP_SIZE = 1000;

    /** warmup time */
    public static final long WARMUP_TIME = 5 * 1000;

    /** test time */
    public static final long TEST_TIME = 10 * 1000;
    
    /** percent of throughput by which a race must beat 
     * the direct path for switching to occur */
    public static final double MIN_RACE_PERCENT = 0.05;

    /** number of seconds between successive reads to DHT for peer */
    private static final int DHT_READ_INTERVAL = 60;
    private static final int MAX_DHT_FAILURES = 1;

    /** the maximum number of edges to send to source */
    private static final int MAX_EDGES = 10;
	public static final boolean DO_TRACEROUTE = true;
	/** percent of bandwidth to drop to be considered a serious fault */
	public static final double SERIOUS_DROP = 1-0.33;
    private Random r;

    /** list of current DataTransferConfigurations */
    private ArrayList<DataTransferConfig> experiments;



    /** map of InetAddress to experiment id for source->middle node pending connections */
    private HashMap<InetAddress, Integer> pendingMiddleConnections;

    /** map of experiment id to set of middle nodes that are ready to race (have TCP connections established) */
    private HashMap<Integer, LinkedList<IOnoConnection>> readyToRace;


    /** set of destinations for middle hops */
    private HashSet<InetAddress> middleHopDests;

    /** map of socket address to a list of pending messages */
    private HashMap<InetAddress, ArrayList<ByteBuffer>> pendingPackets;

    /** map of destination IP address to experiment
     * Only contains entries for which this node is a source
     *  */
    private HashMap<String, DataTransferConfig> myDestinations;

    /** map of destination to edges for mapping */
    private HashMap<String, ArrayList<String>> destinationEdges;

    /** set of sources of data coming to me */
    private HashMap<String, DataTransferConfig> sources;

    /** vector of queued experiments */
    private Vector<DataTransferConfig> queuedExperiments;

    /** hard coded other peers */
    private Vector<InetAddress> otherPeers;

    /** my IP address */
    private String myIp;

    /** my InetAddress */
    private InetAddress myInetAddress;

    /** maps experiment number to a map between a peer and a set of pairs of values, the
     * pairs being the number of bytes received and the time of receipt
     * for each message */
    private HashMap<Integer, HashMap<InetAddress, Vector<Pair<Long, Long>>>> timingData;

    /** maps an experiment number of a set of paris of values, the
     * pairs being the number of bytes received and the time of receipt
     * for each PRIMARY PATH data message */
    private HashMap<Integer, Vector<Pair<Long, Long>>> entireDataTransferStats;

    /** Map of experiment id to a middle peer's socket address and the value (in ms) of the
     * last time it was probed */
    private HashMap<Integer, LinkedHashMap<InetSocketAddress, Long>> middleNodes;

    /** Map of current edges to set of peers */
    private HashMap<String, HashSet<InetSocketAddress>> peersPerEdge;

    /** Map of all edges (class C subnets) to set of peers */
    private HashMap<String, HashSet<InetSocketAddress>> allPeersPerEdge;
    

    /** map of experiment id to stream handler for the expeirment*/
    private HashMap<Integer, StreamHandler> streamHandlers;

    /** maps a middle hop peer to the index in the total data transfer to compare
     * against. */
    private HashMap<InetAddress, Integer> raceIndexes = new HashMap<InetAddress, Integer>();;

    /** maps an experiment id to the current primary source for the experiment */
    private HashMap<Integer, InetAddress> currentPrimary;
    private HashMap<Integer, Long> primaryBytes = new HashMap<Integer, Long>();

    /** maps experiment id to edge server timer  */
    private HashMap<Integer, ITimer> timers;
    private HashMap<Integer, EdgeServerListHandler> edgeListHandlers;

    /** maps a DHT key to a number of failures that it has had */
    private HashMap<String, Integer> dhtFailures;
    
    /** maps ip address to number of times the connection has failed */
    private HashMap<String, Integer> retries;

    /** maps edge server ip (class C subnet) to a list of hard-coded peer ips that are mapped to them */
    private HashMap<String, HashSet<String>> hardCodedMappings;
	private IDistributedDatabase ddb;
	private IOnoConnectionManager manager;
	public boolean shouldSendPeerList;
	public Set<InetSocketAddress> peersCloseToMe;
	private ITimer timer;
	
	/** map of transfer object to socket */
	private HashMap<Object, InetSocketAddress> transfers;
	/** map of transfer obejct to config object */
	private HashMap<DataTransferConfig, Object> configToTransfer;
	private HashMap<Object, DataTransferConfig > transferToConfig;
	private HashMap<Object, InputStream > destInputStreams;	
	private HashMap<Object, OutputStream > destOutputStreams;

    /**
     *
     */
    public OnoDataTransferManager() {
        super();
        experiments = new ArrayList<DataTransferConfig>();
        //      this.gmr = Main.getGMR();
        pi = MainGeneric.getPluginInterface();
        r = new Random();
        sources = new HashMap<String, DataTransferConfig>();
        myDestinations = new HashMap<String, DataTransferConfig>();
        timingData = new HashMap<Integer, HashMap<InetAddress, Vector<Pair<Long, Long>>>>();
        entireDataTransferStats = new HashMap<Integer, Vector<Pair<Long, Long>>>();
        
        middleHopDests = new HashSet<InetAddress>();
        pendingPackets = new HashMap<InetAddress, ArrayList<ByteBuffer>>();
        destinationEdges = new HashMap<String, ArrayList<String>>();
        middleNodes = new HashMap<Integer, LinkedHashMap<InetSocketAddress, Long>>();
        peersPerEdge = new HashMap<String, HashSet<InetSocketAddress>>();
        allPeersPerEdge = new HashMap<String, HashSet<InetSocketAddress>>();
        readyToRace = new HashMap<Integer, LinkedList<IOnoConnection>>();
        pendingMiddleConnections = new HashMap<InetAddress, Integer>();
        streamHandlers = new HashMap<Integer, StreamHandler>();
        
        currentPrimary = new HashMap<Integer, InetAddress>();
        timers = new HashMap<Integer, ITimer>();
        edgeListHandlers = new HashMap<Integer, EdgeServerListHandler>();
        dhtFailures = new HashMap<String, Integer>();
        queuedExperiments = new Vector<DataTransferConfig>();
        hardCodedMappings = new HashMap<String, HashSet<String>>();
        otherPeers = new Vector<InetAddress>();
        retries = new HashMap<String, Integer>();
        
        transfers = new HashMap<Object, InetSocketAddress>();
        configToTransfer = new HashMap<DataTransferConfig, Object>();
        transferToConfig = new HashMap<Object, DataTransferConfig >();
         destInputStreams = new HashMap<Object, InputStream >();
    	destOutputStreams = new HashMap<Object, OutputStream > ();
    	
    	try {
            MainGeneric.getConnectionManager().registerManagerListener("ONO_DATA_MSG",
                    "Ono data message",
                    this);
        } catch (IMessageException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }




    /* (non-Javadoc)
     * @see edu.northwestern.ono.experiment.IExperimentRunner#beginExperiment(edu.northwestern.ono.experiment.DataTransferConfiguration)
     */
    public void beginTransfer(DataTransferConfig onoExp) {
        try {
        	if (!onoExp.dest.equals(MainGeneric.getPublicIpAddress()))
        		Thread.sleep(5 * 1000); // sleep for a minute to warm up

            while ((MainGeneric.getPublicIpAddress() == null) ||
                    (MainGeneric.getDistributedDatabase() == null)) {
                Thread.sleep(5 * 1000);
            } // wait for a valid address
            
            ddb = MainGeneric.getDistributedDatabase();
            
            CDNClusterFinder.getInstance().setPreferredName(onoExp.preferredName);

            myIp = Digger.getInstance().getIp();
            myInetAddress = InetAddress.getByName(myIp);

            System.out.println("My ip: " + myIp + "\tSource ip:" +
                onoExp.source);

            // check if this guy is the source or dest
            if (onoExp.dest.equals(myIp)) {
            	
            } else if (onoExp.source.equals(myIp)) {
                //MainGeneric.getHardCodedMappings(onoExp, hardCodedMappings, otherPeers);

                int port = getPort(onoExp);

                //              if (port==-1){
                //              // try the default port for PL nodes
                //              port = 7000;
                //              }

                // start experiment by connecting to dest
                InetSocketAddress otherEnd = new InetSocketAddress(onoExp.dest,
                        port);

                
                manager.connect(otherEnd, this, true);
                
            }
        } catch (InterruptedException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (UnknownHostException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    /**
     * @param myIp2
     * @param onoExp
     */
    private int getPort(DataTransferConfig onoExp) {
        int port = MainGeneric.getPeerPort(onoExp.dest);
        int i = 0;

        if (port == -1) {
            
                return onoExp.port; // give up and try default
            
        }
        else return port;
    }

    /* (non-Javadoc)
     * @see org.gudy.azureus2.plugins.messaging.generic.GenericMessageConnectionListener#connected(org.gudy.azureus2.plugins.messaging.generic.GenericMessageConnection)
     */
    public void connected(IOnoConnection connection) {
    	if (timer!=null){    		
    		timer.destroy();
    	}
        InetSocketAddress sock = connection.getEndpoint().getNotionalAddress();
        String otherEnd = sock.getAddress().toString().substring(1);
        DataTransferConfig onoExp = sources.get(otherEnd);
        
        // this is the source getting an incoming data connection
        if (onoExp!=null && connection.isLocallyIniated()){
        	synchronized(configToTransfer.get(onoExp)){
        		configToTransfer.get(onoExp).notifyAll();
        	}
        	return;
        }

        onoExp = myDestinations.get(otherEnd);
        // TODO deal with all kinds of garbage from middle points
        if (connection.isLocallyIniated() && 
        		(onoExp==null || !otherEnd.equals(onoExp.dest))) {
            // if this is middle hop to dest
            if (isMiddleHop(connection)) {
                while (pendingPackets.get(sock.getAddress()).size() > 0) {
                    synchronized (pendingPackets.get(sock.getAddress())) {
                        try {
                            // send all pending packets
                            connection.send(pendingPackets.get(
                                    sock.getAddress()).get(0));
                        } catch (IMessageException e) {
                            // TODO Auto-generated catch block
                            e.printStackTrace();
                        }

                        pendingPackets.get(sock.getAddress()).remove(0);
                    }
                }
            } // end if for middle hop
            else { // otherwise, this is the source connecting to a middle hop

                // conduct race

                // send a buch of data, then get responses from destination about speed
                // TODO be less trustworthy
                Integer id = pendingMiddleConnections.get(getEndpointInetAddress(
                            connection));

                if (id != null) {
                    synchronized (readyToRace) {
                        if (readyToRace.get(id) == null) {
                            readyToRace.put(id,
                                new LinkedList<IOnoConnection>());
                        }

                        addToCurrentRaces(id, connection);
                    }
                } else {
                    System.err.println(
                        "Middle hop? Unknown connection!");
                    reportError("Middle hop? Unknown connection!");
                }
            }
        
        } else { // this is the source connecting to the dest to send data
            
        }
    }

    /**
    * @param id
    * @param connection
    */
    private void addToCurrentRaces(Integer id,
    		IOnoConnection connection) {
        //        System.out.println("\n\nTrying to add to current race!\nId "+ id + 
        //                "\tCon: "+connection.getEndpoint().getNotionalAddress());
        LinkedList<IOnoConnection> myReadyNodes = readyToRace.get(id);

        synchronized(readyToRace){
	        for (IOnoConnection con : myReadyNodes) {
	            if (getEndpointInetAddress(con)
	                        .equals(getEndpointInetAddress(connection))) {
	                //                System.out.println("\n\nNot adding because aleady in ready list!");
	                return; // already there, don't add it again.
	            } else if (((streamHandlers.get(id) != null) &&
	                    !streamHandlers.get(id)
	                                       .peerCanBeAdded(getEndpointInetAddress(
	                            connection)))) {
	                //                System.out.println("\n\nNot adding because peer lost race recently!");
	                return; // already there, don't add it again.
	            }
	        }
	
	        readyToRace.get(id).add(connection);
        }
        if (streamHandlers.get(id)!=null) // TODO why is this null?
        	streamHandlers.get(id).allowRaceChecking();
    }

    /**
         * @param connection the connection over which to send data
         * @param size the total number of data bytes to send
         * @param type the type of message being send
     * @param outputBuffer 
         */
    private void sendData(InetAddress destAddress, int size,
        byte type, byte[] dest, int port, ByteBuffer outputBuffer) {
        if (DEBUG)
        	System.out.print("Sending " + size + " bytes to " +
        		destAddress);
        
        IOnoConnection connection = manager.getConnection(destAddress);

        int bytesLeft = size;
        OnoMessageData odm = new OnoMessageData(type, dest, port);
        int numSend = Math.min(OnoMessageData.payloadMaxSize(connection), size);
        if (numSend <=0) return;
        byte[] bytesToSend = new byte[numSend];

        // start sending messages... making sure not to overflow the Q (check exception conditions)
        while (bytesLeft > 0) {
            try {
            	if (outputBuffer==null){
            		r.nextBytes(bytesToSend);
            	} else {
            		outputBuffer.get(bytesToSend);
            	}
                odm.setData(bytesToSend);
                connection.send(odm.toByteBuffer());

                // TODO I probably have to free these buffers...
                bytesLeft -= numSend;
                
                //Thread.sleep(5*1000);
            } catch (IMessageException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
                connection.close();
                
            } 
        } // end while bytes left to send
    }

    /**
     *
     * @param sock the endpoint socket
     * @return
     */
    private boolean isMiddleHop(IOnoConnection connection) {
        if (middleHopDests.remove(connection.getEndpoint().getNotionalAddress().getAddress())) {
            return true;
        } else {
            return false;
        }
    }

    /* (non-Javadoc)
    * @see org.gudy.azureus2.plugins.messaging.generic.GenericMessageConnectionListener#receive(org.gudy.azureus2.plugins.messaging.generic.GenericMessageConnection, org.gudy.azureus2.plugins.utils.PooledByteBuffer)
    */
    public void receive(IOnoConnection connection,
        ByteBuffer message) throws IMessageException {
    	
    	boolean returnToPool = true; // if false, buffer should not be returned to pool because it is being cached...
    	// TODO handle reclaiming memory on broken connections.
    	
        // TODO is this ok?
        long receiveTime = System.currentTimeMillis();

        if (myIp == null) {
            if (MainGeneric.getPublicIpAddress() != null) {
                myIp = MainGeneric.getPublicIpAddress();
            }
        }

        // three cases:
        String endpoint = getEndpointIpAddress(connection);
        ByteBuffer bb = message;

        if (bb.remaining() == 0) {
            System.err.println("Why is message empty?");
            reportError("Why is message empty?");
            manager.returnToPool(bb);

            return;
        }
               

        // first byte contains message type            
        byte type = bb.get();
        //        System.out.println("My destinations: "+myDestinations.toString());
        if (DEBUG) System.out.print("Message type: " + type + " (from " + endpoint +
            ")\t");

        if (type == TYPE_PEER_REQUEST) {
            System.out.println("Peer request!");
        }

        if (type == TYPE_EXP_ANNOUNCE){
        	int expId = bb.getInt();
        	if (sources.get(endpoint)==null){
        		// TODO finish        		
        	}
        	manager.returnToPool(bb);
        	return;
        }
        
        byte[] ipBytes = new byte[4];

        // if this message came from an unknown source, see if it came 
        // from a middle hop by checking whether the "dest" address is the source.
        // if so, add this node to the set of sources for this experiment.
        // this is not likely a "done" message
        if ((type == TYPE_PRIMARY_PATH) || (type == TYPE_RACE)) {
            if ((type == TYPE_RACE) && (sources.get(endpoint) == null)) {
                bb.get(ipBytes);
                bb.position(1);

                InetAddress addr;

                try {
                    addr = InetAddress.getByAddress(ipBytes);

                    // TODO this doesn't handle a node being a middle point 
                    // for multiple concurrent experiments to the same dest
                    if (sources.containsKey(addr.toString().substring(1))) {
                        sources.put(getEndpointIpAddress(connection),
                            sources.get(addr.toString().substring(1)));
                    }
                } catch (UnknownHostException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }
        }

        // 1) I am the sender, getting a message from receiver
        if (myDestinations.containsKey(endpoint)) { // mesage from receiver...
                                                    // get experiment object

            DataTransferConfig thisExp = myDestinations.get(endpoint);

            if (streamHandlers.get(thisExp.experimentId) == null) {
                manager.returnToPool(bb);

                return;
            }

            switch (type) {
            // a message w/ a list of edges
            case TYPE_EDGES:
            	System.err.println("Message type \"EDGES\" sould not be sent!");
            	break;
            	/*
                int numEdges = bb.getInt();
                ArrayList<String> oldEdges = destinationEdges.get(endpoint);

                // TODO for now, each time receiver sends such a message, use only the edges 
                // from that message (i.e., wipe out old ones...
                //if (destinationEdges.get(endpoint)==null){
                ArrayList<String> edges = new ArrayList<String>();
                destinationEdges.put(endpoint, edges);

                String newEdge;

                //}
                boolean newEdgeSeen = false;

                for (int i = 0; i < numEdges; i++) {
                    bb.get(ipBytes); // TODO debug and check that this is doing the right thing w/ position...

                    try {
                        InetAddress ipAddress = InetAddress.getByAddress(ipBytes);
                        newEdge = ipAddress.toString().substring(1);
                        edges.add(newEdge);

                        if (!peersPerEdge.containsKey(newEdge)) {
                            System.out.println("New edge seen: " + newEdge);

                            if (allPeersPerEdge.containsKey(
                                        Util.getClassCSubnet(newEdge))) {
                                peersPerEdge.put(newEdge,
                                    allPeersPerEdge.get(Util.getClassCSubnet(
                                            newEdge)));

                                if (middleNodes.get(thisExp.experimentId) == null) {
                                    middleNodes.put(thisExp.experimentId,
                                        new LinkedHashMap<InetSocketAddress, Long>());
                                }

                                for (InetSocketAddress sock : peersPerEdge.get(
                                        newEdge)) {
                                    if (middleNodes.get(thisExp.experimentId)
                                                       .get(sock) == null) {
                                        middleNodes.get(thisExp.experimentId)
                                                   .put(sock, -1L);
                                    }
                                }
                            } else {
                                peersPerEdge.put(newEdge,
                                    new HashSet<InetSocketAddress>());
                            }

                            newEdgeSeen = true;
                        }
                    } catch (UnknownHostException e) {
                        e.printStackTrace();
                    }
                } // end processing of edges

                // now remove peers that are associated w/ dead edges
                if (oldEdges != null) {
                    for (String edge : oldEdges) {
                        if (!edges.contains(edge)) {
                            HashSet<InetSocketAddress> peers = peersPerEdge.remove(edge);

                            if (middleNodes.get(thisExp.experimentId) != null) {
                                synchronized (middleNodes) {
                                    for (InetSocketAddress peer : peers) {
                                        middleNodes.get(thisExp.experimentId)
                                                   .remove(peer);
                                    }
                                }
                            }

                            if (allPeersPerEdge.get(Util.getClassCSubnet(edge)) == null) {
                                allPeersPerEdge.put(Util.getClassCSubnet(edge),
                                    peers);
                            } else {
                                allPeersPerEdge.get(Util.getClassCSubnet(edge))
                                               .addAll(peers);
                            }
                        }
                    }
                }

                if (hardCodedMappings != null) {
                    for (String edge : edges) {
                        if ((hardCodedMappings.get(Util.getClassCSubnet(edge)) != null) &&
                                (hardCodedMappings.get(Util.getClassCSubnet(
                                        edge)).size() > 0)) {
                            for (String peer : hardCodedMappings.get(
                                    Util.getClassCSubnet(edge))) {
                                try {
                                    if (peer == null) {
                                        continue;
                                    }

                                    InetAddress addr = Address.getByAddress(peer);
                                    int port = thisExp.port;
                                    InetSocketAddress sock = new InetSocketAddress(addr,
                                            port);

                                    if (middleNodes.get(thisExp.experimentId) == null) {
                                        middleNodes.put(thisExp.experimentId,
                                            new LinkedHashMap<InetSocketAddress, Long>());
                                    }

                                    if (middleNodes.get(thisExp.experimentId)
                                                       .get(sock) == null) {
                                        middleNodes.get(thisExp.experimentId)
                                                   .put(sock, -1L);
                                    }

                                    if (peersPerEdge.get(edge) == null) {
                                        peersPerEdge.put(edge,
                                            new HashSet<InetSocketAddress>());
                                    }

                                    peersPerEdge.get(edge).add(sock);
                                } catch (UnknownHostException e) {
                                    // TODO Auto-generated catch block
                                    e.printStackTrace();
                                }
                            }
                        }
                    }
                }

                // if we found a new edge, allow races
                // TODO this has to be configurable
                if (newEdgeSeen) {
                    streamHandlers.get(thisExp.experimentId).allowRaceChecking();
                }

                // now do reads
                for (String edge : edges) {
                	ddb.read(edge, String.valueOf(thisExp.experimentId), 15, 
                			this,
                        ddb.getOptions().getExhaustiveRead() |
                        ddb.getOptions().getHighPriority()); // only wait for 15 secs the first time
                }

                // record event
                Statistics.getInstance()
                          .addDataTransferEvent(thisExp.source, myIp,
                    thisExp.dest, thisExp.experimentId, bb.limit(),
                    receiveTime, type, -1.0, -1.0);
			
                break; */

            case TYPE_RACE_RESULT:

                int numEntries = bb.getInt();
                double[] races = new double[numEntries];

                TreeMap<Double, InetAddress> raceResult = new TreeMap<Double, InetAddress>();

                for (int i = 0; i < numEntries; i++) {
                    // get ip address
                    bb.get(ipBytes);

                    try {
                        InetAddress curAddr = InetAddress.getByAddress(ipBytes);

                        if (curAddr.toString().substring(1).equals(myIp)) {
                            curAddr = Address.getByAddress(thisExp.dest);
                        }

                        races[i] = bb.getDouble();
                        raceResult.put(races[i], curAddr);
                    } catch (UnknownHostException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }
                }

                // go with best hop
                Iterator<Entry<Double, InetAddress>> it = raceResult.entrySet()
                                                                    .iterator();
                InetAddress lastOne = null;

                double currentThroughput = 0;
                double bestTP = 0;
                Entry<Double, InetAddress> ent;
                InetAddress currentAddress = streamHandlers.
    			get(thisExp.experimentId).currentInetAddress;
                while (it.hasNext()){
                	ent = it.next();
                	if (ent.getValue().equals( currentAddress)){
                		currentThroughput = ent.getKey();
                	}
                    lastOne = ent.getValue();
                    bestTP = ent.getKey();
                }

                // keep the current path if new throughput is not significantly
                // higher than the current one.
                if (!currentAddress.equals(lastOne) 
                		&& bestTP <= (1+MIN_RACE_PERCENT)*currentThroughput){
                	lastOne = currentAddress;
                }
                
                streamHandlers.get(thisExp.experimentId).setCurrentPathExt(lastOne, 
                		raceResult.entrySet().iterator().next().getValue());

                //            	 record event
                Statistics.getInstance()
                          .addDataTransferEvent(thisExp.source, myIp,
                    thisExp.dest, thisExp.experimentId, bb.limit(),
                    receiveTime, type, races[0], races[1]);

                break;

            case TYPE_PEER_LIST:
                numEntries = bb.getInt();

                int experimentId = thisExp.experimentId;

                for (int i = 0; i < numEntries; i++) {
                    bb.get(ipBytes);

                    try {
                        InetAddress ina = InetAddress.getByAddress(ipBytes);
                        int port = bb.getInt();
                        bb.get(ipBytes);

//                        InetAddress edge = InetAddress.getByAddress(ipBytes);

                        InetSocketAddress sock = new InetSocketAddress(ina, port);

                        // don't add if duplicated
                        if ((middleNodes.get(experimentId) != null) &&
                                middleNodes.get(experimentId).containsKey(sock)) {
                            continue;
                        }

                        //                       skip if it's telling me to detour through me
                        if (sock.getAddress().toString().substring(1)
                                    .equals(myIp)) {
                            continue;
                        }

                        LinkedHashMap<InetSocketAddress, Long> peers;

                        synchronized (middleNodes) {
                            peers = middleNodes.get(experimentId);

                            if (peers == null) {
                                peers = new LinkedHashMap<InetSocketAddress, Long>();
                                middleNodes.put(experimentId, peers);
                            }
                        }

//                        synchronized (peersPerEdge) {
//                            if (peersPerEdge.get(edge.toString()) == null) {
//                                peersPerEdge.put(edge.toString(),
//                                    new HashSet<InetSocketAddress>());
//                            }
//
//                            peersPerEdge.get(edge.toString()).add(sock);
//                        }

                        peers.put(sock, -1L);
                        System.out.println("Added new peer: " + sock);
                        streamHandlers.get(thisExp.experimentId)
                                      .allowRaceChecking();
                    } catch (UnknownHostException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }
                } // end for ip addresses
                  //               record event

                Statistics.getInstance()
                          .addDataTransferEvent(thisExp.source, myIp,
                    thisExp.dest, thisExp.experimentId, bb.limit(),
                    receiveTime, type, -1, -1);

                break;

            case TYPE_ACK:

                try {
                    bb.position(1);
                    bb.get(ipBytes);

                    InetAddress ina = InetAddress.getByAddress(ipBytes);

                    if (ina.equals(myInetAddress)) {
                        // then this is a direct connection ack
                        ina = getEndpointInetAddress(connection);
                    }

                    int ackNum = bb.getInt();

                    if ((thisExp != null) &&
                            (streamHandlers.get(thisExp.experimentId) != null)) {
                        streamHandlers.get(thisExp.experimentId)
                                      .recordAck(ina, ackNum);
                    }
                } catch (UnknownHostException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }

                break;

            default:
                System.err.println("Unknown destination message!");
            	reportError("Unknown destination message!");
            }
        }
        // 2) I am a destination, in which case, record data stats
        else if (sources.get(endpoint) != null) {
        	
            DataTransferConfig thisExp = sources.get(endpoint);
            InetAddress endPoint = getEndpointInetAddress(connection);

            InetAddress curPrimaryAddress = currentPrimary.get(thisExp.experimentId);
            
            if (entireDataTransferStats.get(thisExp.experimentId) != null) {
            	if (DEBUG) System.out.println("Received data (currently at: " +
                    entireDataTransferStats.get(thisExp.experimentId).size() +
                    ")");
            }

            // first see if this is a special request
            if (type == TYPE_PEER_REQUEST) {
                //         	 record event
                Statistics.getInstance()
                          .addDataTransferEvent(thisExp.source, myIp,
                    thisExp.dest, thisExp.experimentId, bb.limit(),
                    receiveTime, type, -1, -1);

                sendPeerList(connection);

                manager.returnToPool(bb);

                return;
            }

            synchronized (timingData) {
                // make the record
                Pair<Long, Long> stat = new Pair<Long, Long>((long) message.limit(),
                        receiveTime);

                // first packet received from this node in this experiment
                if (timingData.get(thisExp.experimentId) == null) {
                    timingData.put(thisExp.experimentId,
                        new HashMap<InetAddress, Vector<Pair<Long, Long>>>());
                }

                if (timingData.get(thisExp.experimentId)
                                  .get(endPoint) == null) {
                    timingData.get(thisExp.experimentId)
                              .put(endPoint,
                        new Vector<Pair<Long, Long>>());
                }

                // record timing data
                timingData.get(thisExp.experimentId)
                          .get(endPoint).add(stat);

                switch (type) {
                case TYPE_PRIMARY_PATH:
                    // update current primary
                    currentPrimary.put(thisExp.experimentId,
                        endPoint);
                    if (primaryBytes.get(thisExp.experimentId)==null){
                    	primaryBytes.put(thisExp.experimentId, (long)(bb.limit()-9));
                    } else {
                    	primaryBytes.put(thisExp.experimentId, (long)(bb.limit()-9 + 
                    			primaryBytes.get(thisExp.experimentId)));
                    }

                    if (entireDataTransferStats.get(thisExp.experimentId) == null) {
                        entireDataTransferStats.put(thisExp.experimentId,
                            new Vector<Pair<Long, Long>>());
                    }

                    entireDataTransferStats.get(thisExp.experimentId).add(stat);

                    if (DEBUG) 
                    	System.out.println("Received " +
                        bb.limit()+ " bytes!(#" +
                        timingData.get(thisExp.experimentId)
                                  .get(endPoint).size() +
                        ") from "+endPoint);
                    Statistics.getInstance()
                              .addDataTransferEvent(getEndpointInetAddress(
                            connection).toString().substring(1), myIp,
                        thisExp.dest, thisExp.experimentId, bb.limit(),
                        receiveTime, type, -1, -1);

                    boolean done = false;

                    if (bb.limit() == 17) { // done message is sent

                        String s = "";
                        bb.position(9);

                        for (int i = 0; bb.remaining() > 1; i++) {
                            s += bb.getChar();
                        }

                        if (s.equals("done")) {
                            System.out.println("Cleaning up transfer ("+primaryBytes.get(thisExp.experimentId)+" bytes)!");
                            try {
                            	
								destOutputStreams.get(configToTransfer.get(thisExp)).flush();
								//destInputStreams.get(configToTransfer.get(thisExp)).close();
								destOutputStreams.get(configToTransfer.get(thisExp)).close();
							} catch (IOException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							}
                            
                            cleanUpTransfer(connection);
                            done = true;
                        }
                    }

                    //        		sendAck(endPoint, entireDataTransferStats.get(thisExp.experimentId).size(), 
                    //        				thisExp.source);
//                    if (timingData.get(thisExp.experimentId)
//                                      .get(endPoint)
//                                      .size()>=9999){
//                    	System.out.println("Sending ack of size: "+timingData.get(thisExp.experimentId)
//                                      .get(endPoint)
//                                      .size());
//                    }
                    sendAck(endPoint,
                            timingData.get(thisExp.experimentId)
                                      .get(endPoint)
                                      .size(), thisExp.source);
                    
                    if (!done) {
                    	bb.position(9);
                    	// send to higher layer
                    	if (destOutputStreams.get(configToTransfer.get(thisExp))!=null){
                    		try {
                    			//ByteOrder oldOrder = bb.order();
                    			//bb.order(ByteOrder.LITTLE_ENDIAN);
                    			byte b;
                    			while (bb.remaining()>0){
                    				b = bb.get();
//                    				if (b<0){
//                    					System.err.println("Bad data?");
//                    				}
                    				destOutputStreams.get(configToTransfer.get(thisExp)).write(b);
                    			}
                    			//destOutputStreams.get(configToTransfer.get(thisExp)).flush();
                    			//bb.order(oldOrder);
							} catch (IOException e) {
								Statistics.getInstance().reportError(e);
								e.printStackTrace();
							}
                    	}
                    	
                    	
                        
//                        if (shouldSendPeerList){
//                        	sendPeerList(connection);
//                        }
                    }

                    break;

                case TYPE_RACE:

                	System.out.println("Detour race: Received " +
                            message.limit() + " bytes! ");
                	synchronized (raceIndexes) {
                		// is this new?
                		if (message.limit() >= NON_WARMPUP_SIZE) {

                			if (raceIndexes.get(getEndpointInetAddress(
                					connection)) == null) {
                				if ((thisExp == null) ||
                						(entireDataTransferStats.get(
                								thisExp.experimentId) == null)) {
                					System.out.println("Current exp: " +
                							thisExp.experimentId);
                					System.out.println("EDTS keys: " +
                							entireDataTransferStats.keySet());

                					if (entireDataTransferStats.get(
                							thisExp.experimentId) == null) {
                						entireDataTransferStats.put(thisExp.experimentId,
                								new Vector<Pair<Long, Long>>());
                					}
                				}

                				//        			 if so, record point in overall data transfer
                				raceIndexes.put(getEndpointInetAddress(
                						connection),
                						entireDataTransferStats.get(
                								thisExp.experimentId).size());
                			}

                		}

                		if (isDoneWithMiddle(bb)) {

                			synchronized (entireDataTransferStats){
                				if ( entireDataTransferStats.
                					get(thisExp.experimentId)!=null){
                					InetAddress middleHop = endPoint;

                					// if (middleHop!=null && raceIndexes!=null &&  timingData.get(thisExp.experimentId)!=null){
                					Integer ri = raceIndexes.remove(middleHop);
                					int rii = -1;
                					if (ri!=null) rii = ri.intValue();
                					RaceReporter rr = new RaceReporter(thisExp, middleHop,
                							timingData.get(thisExp.experimentId)
                							.get(middleHop),rii, entireDataTransferStats.
                							get(thisExp.experimentId).size(), curPrimaryAddress);


                					MainGeneric.createThread("RaceReporter", rr);
//              			}
                				} // experiment not already done
                			} // synch	

                		}
                	} // end synchronized raceIndexes
                    
                    if (DEBUG) System.out.println("(#" +
                            timingData.get(thisExp.experimentId)
                                      .get(endPoint).size() +
                            ")");
                    //         	 record event
                    Statistics.getInstance()
                              .addDataTransferEvent(getEndpointInetAddress(
                            connection).toString().substring(1), myIp,
                        thisExp.dest, thisExp.experimentId, bb.limit(),
                        receiveTime, type, -1, -1);

                    sendAck(endPoint,
                        timingData.get(thisExp.experimentId)
                                  .get(endPoint).size(),
                        thisExp.source);

                    break;

                default:
                    System.err.println("Unsupported message type!");
                	reportError("Unsupported message type!");
                }
            }
        }
        //      3) I am a middle hop, in which case, parse the message and forward
        // or I am the destination and 
        else if (sources.get(endpoint) == null) {
            try {
                if (type == 1) {
                    manager.returnToPool(bb);

                    return;
                }

                System.out.println("Forwarding " +
                    message.limit() + " bytes!");

                if ((type != TYPE_RACE) && (type != TYPE_PRIMARY_PATH)) {
                	connection.close();
                    throw new RuntimeException(
                        "Incorrect message type at middle hop!");
                    
                }

                // get ip address of receiver...
                bb.position(1);
                bb.get(ipBytes);

                if ((ipBytes[0] == 0) && (ipBytes[1] == 0) &&
                        (ipBytes[2] == 0) && (ipBytes[3] == 0)) {
//                    throw new RuntimeException(
//                        "Zero IP when should have destination!");
                	System.err.println("Unrecognized data from "+endpoint);
                	reportError("Unrecognized data from "+endpoint);
                	manager.returnToPool(bb);
                    return;
                }

                InetAddress ipAddress = InetAddress.getByAddress(ipBytes);
                int port = bb.getInt();
                InetSocketAddress dest = new InetSocketAddress(ipAddress, port);
                System.out.println("Destination is: " + dest);
                
                

                // port is no good w/ server sockets...
                IOnoConnection con = manager.getConnection(dest.getAddress());

                // update message with ip of source
                bb.position(1);
                bb.put(getEndpointInetAddress(connection).getAddress());

                //bb.position(9);
                bb.position(0);

                // send data if connection already exists...
                if (con != null) {
                    // enqueue if Q is not empty...
                    if (pendingPackets.get(dest.getAddress()) != null && 
                    		pendingPackets.get(dest.getAddress()).size() > 0) {
                        synchronized (pendingPackets.get(dest.getAddress())) {
                            pendingPackets.get(dest.getAddress()).add(bb);
                            returnToPool = false;
                        }
                    } else {
                        // reposition and send the data
                        // TODO do I have to create a new buffer?
                    	bb.position(0);
                        con.send(bb);
                        returnToPool = false;
                    }
                }
                // otherwise, enqueue data and make connection...
                else {
                    if (!manager.hasConnection(dest)) {
                    	manager.connect(dest, this, true);
                        middleHopDests.add(dest.getAddress());
                        pendingPackets.put(dest.getAddress(), new ArrayList<ByteBuffer>());
                        if (DO_TRACEROUTE){
                        	// this allows us to complete the route
                        	TraceRouteRunner.getInstance().addIp(
                        			ipAddress.getHostAddress());
                        }
                    }

                    pendingPackets.get(dest.getAddress()).add(bb);
                    
                    returnToPool = false;
                }

                //         	 record event
                Statistics.getInstance()
                          .addDataTransferEvent(getEndpointIpAddress(connection),
                    myIp, ipAddress.toString().substring(1), -1, bb.limit(),
                    receiveTime, type, -1, -1);

                // finally, check to see if this is the end of the 
                // the data transfer, and if so, commit results
                if (isDoneWithMiddle(bb.duplicate())) {
                    Statistics.getInstance().commitDataTransfers();
                }
            } catch (UnknownHostException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }

        // 4) error
        // System.out.println("Message received: " + new String(message.toByteArray()));
        if (returnToPool) manager.returnToPool(bb);
    }

	private void sendPeerList(IOnoConnection connection) {
		
		 Set<OnoPeer> tempOnoPeersCloseToMe = 
 			CDNClusterFinder.getInstance().getPreferredPeers(
 					MainGeneric.getClusterFinderObject(connection.getEndpoint().getNotionalAddress().getAddress().getHostAddress()));
         Set<InetSocketAddress> peersToSend = new HashSet<InetSocketAddress>();
         for (OnoPeer p : tempOnoPeersCloseToMe){
        	 peersToSend.add(new InetSocketAddress(p.getIp(), p.getPort()));
         }

//                Digger.getInstance()
//                                                                           .getNearbyOnoPeers(MAX_EDGES);

		if (peersToSend.size() > 0) {
		    ByteBuffer bb2 = manager.getByteBuffer(1 + 4 +
		            (peersToSend.size() * 12));
		    bb2.put(TYPE_PEER_LIST);
		    // TODO swap comments if debugging
		    //                    bb2.putInt(1);
		    bb2.putInt(peersToSend.size());

		    boolean send = false;

		    for (InetSocketAddress p : peersToSend) {
//                        try {
		            // TODO remove when done debugging
		            //if (!p.getKey().contains("165")) continue;
		            bb2.put(p.getAddress().getAddress()); // peer ip
		            bb2.putInt(p.getPort()); // peer port
		            bb2.put(new byte[]{0,0,0,0});
//                            		Address.getByAddress(
//                                    p.getValue().getValue()).getAddress()); // peer's edge server
		            System.out.print("\t" + p + "\t");
		            System.out.println();
		            send = true;
//                        } catch (UnknownHostException e) {
//                            // TODO Auto-generated catch block
//                            e.printStackTrace();
//                        }
		    }

		    if (send) {
		        bb2.position(0);
		        connection.send(bb2);
		    }
		}
	}

    /**
    * @param newEdge
    * @return
    * @deprecated Use {@link Util#getClassCSubnet(String)} instead
    */
    private static String getClassCSubnet(String ipaddress) {
        return Util.getClassCSubnet(ipaddress);
    }

    private void sendAck(InetAddress endpointInetAddress, int size,
        String sourceS) {
        //		 get connection for dest
        InetAddress source;

        try {
            source = Address.getByAddress(sourceS);

            for (Entry<InetSocketAddress, IOnoConnection> e : 
            	manager.getActiveConnections().entrySet()) {
                if (e.getKey().getAddress().equals(source)) {
                    ByteBuffer bb = manager.getByteBuffer(9);
                    bb.put(TYPE_ACK);
                    bb.put(endpointInetAddress.getAddress());
                    bb.putInt(size);
                    bb.position(0);

                    e.getValue().send(bb);

                    if (DEBUG) System.out.println("Sent ack to " + e.getKey());
                }
            }
        } catch (UnknownHostException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
        } catch (IMessageException e2) {
            // TODO Auto-generated catch block
            e2.printStackTrace();
        }
    }

    private boolean isDoneWithMiddle(ByteBuffer bb) {
        if (bb.limit() == 17) { // done message is sent

            String s = "";
            bb.position(9);

            for (int i = 0; i < 4; i++)
                s += bb.getChar();

            if (s.equals("done")) {
                return true;
            }
        } else if (bb.limit() == 9) { // done message is sent

            String s = "";
            bb.position(1);

            for (int i = 0; i < 4; i++)
                s += bb.getChar();

            if (s.equals("done")) {
                return true;
            }
        }

        return false;
    }



    /**
     * @param connection
     */
    private void cleanUpTransfer(IOnoConnection connection) {
        DataTransferConfig onoExp = sources.remove(getEndpointIpAddress(connection));

        if (onoExp != null) {
            // first try to write stats to database
            Statistics.getInstance().commitDataTransfers();

            if (timers.get(onoExp.experimentId) != null) {
                timers.get(onoExp.experimentId).destroy();
                edgeListHandlers.remove(onoExp.experimentId).disable();
                timers.remove(onoExp.experimentId);
            }

            Vector<Pair<Long, Long>> data = entireDataTransferStats.get(onoExp.experimentId);
            FileOutputStream out; // declare a file output object
            PrintStream p; // declare a print stream object

            try {
                String file = pi.getPluginDirectoryName() + File.separatorChar +
                    "StreamData" + onoExp.experimentId + ".dat";
                // Create a new file output stream
                // connected to "myfile.txt"
                out = new FileOutputStream(file);

                // Connect print stream to the output stream
                p = new PrintStream(out);

                p.print(data.toString());

                p.close();
            } catch (Exception e) {
                System.err.println("Error writing to file");
            }

            experiments.remove(onoExp);

            synchronized (timingData) {
                timingData.get(onoExp.experimentId).clear();
            }

            peersPerEdge.clear();
            streamHandlers.remove(onoExp.experimentId);
            synchronized(middleNodes){middleNodes.remove(onoExp.experimentId);}
            synchronized(entireDataTransferStats) {
            	entireDataTransferStats.remove(onoExp.experimentId);
            }
            connection.removeListener(this);

            try {
                connection.close();

                for (IOnoConnection con : manager.getActiveConnections().values()) {
                    con.close();
                }

//                connections.clear();
            } catch (IMessageException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }

    }



    /* (non-Javadoc)
     * @see org.gudy.azureus2.plugins.messaging.generic.GenericMessageConnectionListener#failed(org.gudy.azureus2.plugins.messaging.generic.GenericMessageConnection, java.lang.Throwable)
     */
    public void failed(IOnoConnection connection, Throwable error)
        throws IMessageException {
    	InetSocketAddress sock = null;
    	if (connection.getEndpoint()!=null){
    		sock = connection.getEndpoint().getNotionalAddress();
    		System.out.println("Connection failed to "+sock);
    	}
    	String endpointAddress = getEndpointIpAddress(
                connection);
    	
    	if (sources.containsKey(endpointAddress)){
            DataTransferConfig onoExp = sources.get(endpointAddress);
            if (currentPrimary != null) {
                currentPrimary.remove(onoExp.experimentId);
            }
            
            try {
            	destOutputStreams.get(configToTransfer.get(onoExp)).flush();
            	destOutputStreams.remove(configToTransfer.get(onoExp)).close();
				destInputStreams.remove(configToTransfer.get(onoExp)).close();
				
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
            
            
            if (entireDataTransferStats != null) {
                // only clear if this is the source
                if (sock!=null && endpointAddress.equals(onoExp.source)) {
                    Vector<Pair<Long, Long>> v = 
                    	entireDataTransferStats.get(onoExp.experimentId);

                    if (v != null) {
                        v.clear();
                    }
                }
            }
    	}
    	
        for (DataTransferConfig onoExp : experiments) {
            // reset data transfer
//            if (streamHandlers.get(onoExp.experimentId) != null) {
//                streamHandlers.get(onoExp.experimentId)
//                              .resetAcks(getEndpointInetAddress(connection));
//                streamHandlers.get(onoExp.experimentId).removeRace(connection);
//                streamHandlers.get(onoExp.experimentId).cancelStream();
//            }



            //            if (readyToRace!=null) readyToRace.get(onoExp.experimentId).remove(connection);


//            if (timingData != null) {
//                synchronized (timingData) {
//                    timingData.clear();
//                } // TODO figure out why the statement below does not work                
//                  // timingData.remove(onoExp.experimentId);
//            }

            //            if (raceIndexes!=null) raceIndexes.clear();
            removeConnection(onoExp.experimentId, connection);

            // find connection to remove from races
            if ((readyToRace.get(onoExp.experimentId) != null) &&
                    readyToRace.get(onoExp.experimentId).contains(connection)) {
            	 synchronized(readyToRace){
            		 readyToRace.get(onoExp.experimentId).remove(connection);
            	 }
            }
        }

        DataTransferConfig onoExp = myDestinations.get(endpointAddress);

        if (onoExp != null) {
            if (timingData.get(onoExp.getExperimentId()) != null) {
                synchronized (timingData) {
                    timingData.get(onoExp.getExperimentId()).clear();
                }
            }

            if (entireDataTransferStats.get(onoExp.experimentId) != null) {
                entireDataTransferStats.get(onoExp.experimentId).clear();
            }

            if (timers.get(onoExp.experimentId) != null) {
                timers.remove(onoExp.experimentId).destroy();
                edgeListHandlers.remove(onoExp.experimentId).disable();
            }

            if (streamHandlers.get(onoExp.experimentId) != null) {
                streamHandlers.get(onoExp.experimentId)
                              .resetAcks(getEndpointInetAddress(connection));

                streamHandlers.get(onoExp.experimentId).removeRace(connection);
                if (streamHandlers.get(onoExp.experimentId).isDone()){
                
                	return; // don't reconnect when other end closes connection
                }
            }

            String dest = getEndpointIpAddress(connection);
            int port = getPort(onoExp);

            // try 5 times
            if (retries.get(dest)==null){
            	retries.put(dest, 0);
            }
            if (retries.get(dest)>5){
            	// too many failures
            	System.err.println("Too many failures!");
            	myDestinations.remove(getEndpointIpAddress(
                        connection));
            	experiments.clear();
            	retries.remove(dest);
            }
            else {
            	retries.put(dest, retries.get(dest)+1);
	            final InetSocketAddress otherEnd = new InetSocketAddress(dest, port);
	            final IOnoConnectionManager newManager = manager;
	            final OnoDataTransferManager self = this;
	            timer = MainGeneric.createTimer("ReconnectTimer");
	            timer.addEvent(System.currentTimeMillis()+5*1000, new ITimerEventPerformer(){
	
					public void perform(ITimerEvent event) {
						newManager.connect(otherEnd, self, true);
						
					}});
            }

            return;
        }

        // middle node
        if ((sources.size() == 0) && (myDestinations.size() == 0)) {
            Statistics.getInstance().commitDataTransfers();
        }
    }

    private void removeConnection(int experimentId,
        IOnoConnection connection) {
    	synchronized(middleNodes){
	        if (middleNodes.get(experimentId) != null) {
	            InetSocketAddress toRemove = null;
	            HashMap<InetSocketAddress, Long> vals = middleNodes.get(experimentId);
	
	            for (InetSocketAddress sock : vals.keySet()) {
	                if (sock.getAddress()
	                            .equals(connection.getEndpoint().getNotionalAddress()
	                                                  .getAddress())) {
	                    toRemove = sock;
	
	                    break;
	                }
	            }
	
	            vals.remove(toRemove);
	        }
    	}
    }

    /**
    * @return
    */
    public static OnoDataTransferManager getInstance() {
        if (self == null) {
            self = new OnoDataTransferManager();
        }       
        return self;
    }

    /**
     * @param gmr2
     */
    public void setManager(IOnoConnectionManager manager) {
        this.manager = manager;
    }

    /* (non-Javadoc)
     * @see edu.northwestern.ono.dht.IDHTReadAction#handleRead(byte[], org.gudy.azureus2.plugins.ddb.DistributedDatabaseEvent)
     */
    public void handleRead(byte[] b, IDistributedDatabaseEvent event) {
        // TODO process peers

        // TODO figure out how to measure race
//        try {
            System.out.println("Value read: " + event.getKey().getKey() +
                " -> " + new String(b));
//        } catch (DistributedDatabaseException e1) {
//            // TODO Auto-generated catch block
//            e1.printStackTrace();
//        }

        byte[] bytes = b;
        ByteArrayInputStream bais = new ByteArrayInputStream(bytes);
        int experimentId = Integer.valueOf(event.getKey().getDescription());

        try {
            ObjectInputStream ois = new ObjectInputStream(bais);

            Object o = ois.readObject();

            if ((o instanceof String[])) {
                String[] output = (String[]) o;

                // quit if using old version
                if (Constants.compareVersions(pi.getPluginVersion(), output[0]) < 0) {
                    return;
                }

                // find out which experiment this belongs to...
                DataTransferConfig currentExp = null;

                for (DataTransferConfig exp : experiments) {
                    if (exp.experimentId == Integer.parseInt(
                                event.getKey().getDescription())) {
                        currentExp = exp;

                        break;
                    }
                }

                if (currentExp == null) {
                    throw new RuntimeException(
                        "Cannot match edge to experiment!");
                }

                // match against *OnoExp* torrent
                // get list of compatible peers...
                HashMap<String, Integer> peerIps = MainGeneric.getOnoPeers(100);

                for (String ipMappedToEdge : output) {
                    if (ipMappedToEdge.charAt(0) == '/') {
                        ipMappedToEdge = ipMappedToEdge.substring(1);
                    }

                    if (org.xbill.DNS.Address.isDottedQuad(ipMappedToEdge)) {
                        int port = currentExp.port;

                        // does he match one in the peer list?
                        if (peerIps.containsKey(ipMappedToEdge)) {
                            port = peerIps.get(ipMappedToEdge);
                        }

                        // yes he does, so add him to list of peers for detouring
                        // and begin race
                        InetAddress ina = Address.getByAddress(ipMappedToEdge);

                        InetSocketAddress sock = new InetSocketAddress(ina, port);

                        // don't add if duplicated
                        if ((middleNodes.get(experimentId) != null) &&
                                middleNodes.get(experimentId).containsKey(sock)) {
                            continue;
                        }

                        // skip if it's telling me to detour through me
                        if (sock.getAddress().toString().substring(1)
                                    .equals(myIp)) {
                            continue;
                        }

                        LinkedHashMap peers;

                        synchronized (middleNodes) {
                            peers = middleNodes.get(experimentId);

                            if (peers == null) {
                                peers = new LinkedHashMap<InetSocketAddress, Long>();
                                middleNodes.put(experimentId, peers);
                            }
                        }

                        peers.put(sock, -1L);

                        if (peersPerEdge.get((String) event.getKey().getKey()) == null) {
                            continue;
                        }

                        // otherwise, continue looking
                        synchronized (peersPerEdge.get((String) event.getKey()
                                                                     .getKey())) {
                            peersPerEdge.get((String) event.getKey().getKey())
                                        .add(sock);
                        }

                        // end if in peer list
                    } // end if is valid
                } // end for peers in DHT
            }

            // wait before reading again
            Thread.sleep(DHT_READ_INTERVAL * 1000);
            // while we're here, issue another read
           ddb.read((String) event.getKey().getKey(),
                event.getKey().getDescription(), 60, this,
                ddb.getOptions().getExhaustiveRead() |
                ddb.getOptions().getHighPriority());
        } catch (IOException e) {
            System.err.println(e.toString());

            // generally a corrupted stream and i have to figure out why
            //e.printStackTrace();
        } catch (ClassNotFoundException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
//        } catch (DistributedDatabaseException e) {
//            // TODO Auto-generated catch block
//            e.printStackTrace();
        } catch (InterruptedException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } finally {
            //peer_mon.exit();
        }
    }

    /* (non-Javadoc)
     * @see edu.northwestern.ono.dht.IDHTReadAction#handleTimeout(org.gudy.azureus2.plugins.ddb.DistributedDatabaseEvent)
     */
    public void handleTimeout(IDistributedDatabaseEvent event) {
        try {
            System.out.println("Timeout for reading " +
                event.getKey().getKey() + "!");

            Integer failures = dhtFailures.get(event.getKey().getKey());

            if (failures == null) {
                failures = 0;
            }

            failures = failures + 1;
            dhtFailures.put((String) event.getKey().getKey(), failures);

            if ((failures + 1) > MAX_DHT_FAILURES) {
                // send direct request
            	manager.send(new byte[]{TYPE_PEER_REQUEST}, 
            			getDirectIp(Integer.parseInt(
                            event.getKey().getDescription())));
                
            } else {
                System.out.print("\tSuppressing request: " + failures +
                    " failures...");
            }

            ddb.read((String) event.getKey().getKey(),
                event.getKey().getDescription(), 60, this,
                ddb.getOptions().getExhaustiveRead() |
                ddb.getOptions().getHighPriority());
//        } catch (DistributedDatabaseException e) {
//            // TODO Auto-generated catch block
//            e.printStackTrace();
        } catch (IMessageException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    

    private String getDirectIp(int i) {
    	for (DataTransferConfig exp : experiments) {
            if (exp.experimentId == i) {
                return exp.dest;
            }
        }

        return null;
	}

	/* (non-Javadoc)
    * @see edu.northwestern.ono.dht.IDHTReadAction#handleComplete(org.gudy.azureus2.plugins.ddb.DistributedDatabaseEvent)
    */
    public void handleComplete(IDistributedDatabaseEvent event) {
        if (event.getValue() == null) {
//            try {
                ddb.read((String) event.getKey().getKey(), event.getKey().getDescription(), 30, this, ddb.getOptions().getExhaustiveRead() |
                        ddb.getOptions().getHighPriority());
//            } catch (DistributedDatabaseException e) {
//                // TODO Auto-generated catch block
//                e.printStackTrace();
//            }
        }
    }



    /**
     * @param edgeServerListTimer
     * @param handler
     */
    public boolean setEdgeServerListTimer(IOnoConnection con,
    		ITimer edgeServerListTimer, EdgeServerListHandler handler) {
        DataTransferConfig onoExp = sources.get(getEndpointIpAddress(con));

        if (onoExp != null) {
            System.out.println("Setting edge server list timer!");

            timers.put(onoExp.experimentId, edgeServerListTimer);
            edgeListHandlers.put(onoExp.experimentId, handler);

            return true;
        } else {
            System.out.println("NOT setting edge server list timer!");

            return false;
        }
    }

    /**
     *
     * The EdgeServerListHandler class periodically sends a list of edge
     * servers from the destination to the source.
     */
    public class EdgeServerListHandler implements ITimerEventPerformer {
        /** the connection to the source */
        private IOnoConnection connection;

        /** controls whether to quit */
        private boolean running = true;

        /**
         * @param connection
         */
        public EdgeServerListHandler(IOnoConnection connection) {
            this.connection = connection;
        }
        
        public EdgeServerListHandler() {

        }

        /* (non-Javadoc)
         * @see org.gudy.azureus2.plugins.utils.UTTimerEventPerformer#perform(org.gudy.azureus2.plugins.utils.UTTimerEvent)
         */
        public void perform(ITimerEvent event) {
            if (!running) {
                return;
            }
            
            Set<OnoPeer> tempOnoPeersCloseToMe = 
    			CDNClusterFinder.getInstance().getPreferredPeers(
    					MainGeneric.getClusterFinderObject(null));
            Set<InetSocketAddress> tempPeersCloseToMe = new HashSet<InetSocketAddress>();
            for (OnoPeer p : tempOnoPeersCloseToMe){
            	tempPeersCloseToMe.add(new InetSocketAddress(p.getIp(), p.getPort()));
            }
            if (StreamHandler.HACK_MIDDLE){
            	if (tempPeersCloseToMe == null) return;            	
            }
            peersCloseToMe = tempPeersCloseToMe;
            /*

            String myIp = MainGeneric.getPublicIpAddress();
//            LinkedHashSet<String> myEdges = Digger.getInstance()
//                                                  .getEdgeServerCache().get(myIp);
            
            Object key = MainGeneric.getClusterFinderObject();
            if (key==null) return;
            
            
            Set<String> preferredPeers = null;
//            if (CDNClusterFinder.getInstance().getPreferredPeers(key)!=null){
//            	// TODO add logic to check for NEW paths
//            	shouldSendPeerList = true;
//            }
            
            if (preferredPeers == null) return;

//            int size = preferredPeers.size();
            int i = 0;
            ArrayList<InetAddress> edgesToSend = new ArrayList<InetAddress>();

            // grab the most recent edges
            synchronized (preferredPeers) {
                if (preferredPeers == null) {
                    return; // should not happen...
                            // ones we want are at end of list...
                }

                Iterator<String> it = preferredPeers.iterator();

                // this list was given in reverse order, so when adding, always add to front
                while (it.hasNext()) {
                    if (i < DataTransferRunner.MAX_EDGES) {
                        try {
                            edgesToSend.add(Address.getByName(it.next()));
                        } catch (UnknownHostException e) {
                            // TODO Auto-generated catch block
                            e.printStackTrace();
                        }
                    } // end if last edges
                    else {
                        break;
                    }

                    i++;
                }
            }

            if (edgesToSend.size() == 0) {
                return;
            }

            // **************************************
            // TODO i accidentally converted this to send nearby peers, but 
            // we want to send our ratio map!
            
            
            // now that we have edges, 
            // convert list into bytes and send
            int messageSize = 1 + 4 + (4 * edgesToSend.size());
            ByteBuffer bb = manager.getByteBuffer(messageSize);
            bb.put(DataTransferRunner.TYPE_EDGES); // TODO change message type
            bb.putInt(edgesToSend.size());

            for (InetAddress addr : edgesToSend)
                bb.put(addr.getAddress());

            // always have to reposition the byte buffer
            bb.position(0);

            // send message
            try {
               // TODO  connection.send(bb);
            } catch (IMessageException e) {
               
                e.printStackTrace();
            }
            */
        }

        /**
         * Disables this timer.
         */
        public void disable() {
            running = false;
            // TODO cancel the actual timer
        }
    }

    /**
     *
     * The RaceReporter class takes the results of a
     * race, calculates bandwidth and sends that data
     * back to the destination.
     */
    public class RaceReporter implements Runnable {
        DataTransferConfig myExp;
        InetAddress middleHop;
        Vector<Pair<Long, Long>> raceData;
        int raceIndex;
		private InetAddress currentPrimary;
        private int currentIndex;

        /**
         *
         * @param myExp the current experiment configuration
         * @param middleHop middle hop address for this race
         * @param raceData data for the race, as pairs of bytes and timestamps
         * @param raceIndex the index into the overall data transfer when the race started
         * @param curPrimaryAddress 
         */
        public RaceReporter(DataTransferConfig myExp, InetAddress middleHop,
            Vector<Pair<Long, Long>> raceData, int raceIndex, int currentIndex,
            InetAddress curPrimaryAddress) {
            this.myExp = myExp;
            this.middleHop = middleHop;
            this.raceData = raceData;
            this.raceIndex = raceIndex;
            this.currentIndex = currentIndex;
            this.currentPrimary = curPrimaryAddress;
        }

        /**
         * The main functionality of the race reproter
         */
        public void run() {
            try {
            	
                InetAddress primaryAddr = currentPrimary;
                if (primaryAddr == null){
                	System.err.println("Null primary!");
                	reportError("Null primary in RaceReporter!");
                }

                // get index for comparison w/ primary path
                // TODO fix null pointer here...
                int indexIntoPrimary = raceIndex;

                // get secondary bandwidth
                double detourRate = raceIndex>=0 ? getRate(raceData, 0, raceData.size()) : 0;

                // get bps transferred over primary
                Vector<Pair<Long, Long>> primaryData = entireDataTransferStats.get(myExp.experimentId);

                if (primaryData == null) {
                    System.out.println("My experiment: " + myExp.experimentId);
                    System.out.println("Entire data transfer keys: " +
                        entireDataTransferStats.keySet());
                }

                if (indexIntoPrimary > primaryData.size()) {
                    System.out.println("Overflow!");
                }

                // get primary bandwidth
                double directRate = getRate(primaryData, indexIntoPrimary, currentIndex);

                // record diff
                System.out.println("Pimary rate: " + directRate +
                    "\tRace rate: " + detourRate);

                // get connection for dest
                InetAddress source = Address.getByAddress(myExp.source);

                for (Entry<InetSocketAddress, IOnoConnection> e : manager.getActiveConnections().entrySet()) {
                    if (e.getKey().getAddress().equals(source)) {
                        // report result to dest
              
                        ByteBuffer bb = manager.getByteBuffer(29);
                        bb.put(TYPE_RACE_RESULT);
                        bb.putInt(2);
                        bb.put(middleHop.getAddress());
                        bb.putDouble(detourRate);
                        // TODO pass as argument so it doesn't change under us
                        bb.put(primaryAddr.getAddress());
                        bb.putDouble(directRate);
                        bb.position(0);
                        e.getValue().send(bb);
                    }
                }

                // remove the data for the middle hop
                synchronized (timingData) {
                    timingData.get(myExp.experimentId).remove(middleHop);
                

                int size = timingData.get(myExp.experimentId).get(primaryAddr)!=null ? timingData.get(myExp.experimentId).get(primaryAddr)
                                     .size() : 0;

                // sleep to allow queued packets to arrive if path is switching...
                if (directRate < detourRate) {
                    Thread.sleep(15 * 1000);
                }

                // reset primary, too
                if (size>0 && timingData.get(myExp.experimentId).get(primaryAddr)!=null && 
                		timingData.get(myExp.experimentId).remove(primaryAddr).size() > size) {
                    sendAck(primaryAddr, 0, myExp.source);
                }
                }
            } catch (UnknownHostException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            } catch (IMessageException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            } catch (InterruptedException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }

        /**
         * Gets the bandwidth during the transfer.
         * @param data the packets received
         * @param startIndex the first packet to look at
         * @param endIndex 
         * @return the bandwidth in KB/s
         */
        private double getRate(Vector<Pair<Long, Long>> data, int startIndex, int endIndex) {
            int dataSum = 0;
            int timeIndex = ((startIndex > 0) ? (startIndex - 1) : 0);
            long lastTime = -1;
            int timeSum = 0;
            if (startIndex==-1) return 0;

            for (int i = startIndex; i < endIndex; i++) {
                // handle primary data packets
                if ((timeIndex >= 0) &&
                        (data.get(i).getKey().intValue() > (WARMUP_SIZE + 9))) {
                    dataSum += data.get(i).getKey().intValue();
                    timeSum += data.get(i).getValue()-lastTime;
                } else {
                    // handle race data packets
                    if (data.get(i).getKey().intValue() == (WARMUP_SIZE + 9)) {
//                        System.out.println("Size: " +
//                            data.get(i).getKey().intValue());
//                        System.out.println("Rate index: " + i);
                        timeIndex = i + 1;
                    } else { // otherwise this was a warmup packet and should not be counted
//                        System.out.println("Not counting: " +
//                            data.get(i).getKey().intValue());
                    }
                    
                }
                lastTime = data.get(i).getValue();
            }

            if (data.size() <= startIndex) {
            	String errString = "Incorrect index size: \n" + "Data size: " +
                data.size() + "\tIndex value: " + startIndex;
                System.err.println(errString);
//                reportError(errString);

                return 0;
            }

            // find out how much time elapsed
//            long time = data.get(data.size() - 1).getValue() -
//                data.get(timeIndex).getValue();

            // return the value in KB/s
            double ret = ((double) dataSum) / ((lastTime-data.get(timeIndex).getValue()) / 1E3);

            if (Double.isInfinite(ret)) {
                System.err.println("Invalid value!");
            }

            if (Double.isNaN(ret)) {
                System.err.println("Rate is NaN!");
                ret = -1;
            }

            return ret;
        }
    }

    /**
     *
    * The StreamHandler class manages sending one or more streams of data
    * that are potentially forwarded through detour nodes.
     */
    public class StreamHandler implements Runnable {
        /** time between testing of path for a peer */
        private static final long PEER_PROBE_INTERVAL = 3 * 60 * 1000;

        /** time between testing path to ANY peer */
        private static final long ANY_PROBE_INTERVAL = 1 * 60 * 1000;

        /** number of packets to send ahead of direct when racing */
        private static final int INTERVAL = 4;

        /** debug constant */
        private static final boolean STREAM_DEBUG = false;

        /** frequency with which the list of "bad paths" is cleared */
        private static final long SCORECARD_REFRESH = 10 * 60 * 1000;

        /** if true, uses a hard-coded value for the middle node */
        private static final boolean HACK_MIDDLE = false;
        
        /** number of seconds to wait before transfer times out */
        private static final int TIMEOUT = 60;
        
        /** maximum number of flows to race at once */
        private static final int MAX_CONCURRENT_RACES = 2;
        
        /** throughput window for observation (in seconds) */
        private static final int WINDOW_SIZE = 5;
        
        public boolean listenerActivated = false;

        /** the current experiment */
        private DataTransferConfig onoExp;

        /** the direct connection between source and destination */
        private IOnoConnection directConnection;

        /** string representation of the destinatoin's IP address */
        private String otherEnd;

        /** the data transfer object to which this handler belongs */
        private OnoDataTransferManager parent;

        /** the current primary connection for sending bytes */
        private IOnoConnection currentConnection;

        /** map of connection to number of bytes left send over the connection */
        private HashMap<IOnoConnection, Long> myRaces;

        /** map of InetAddress to last ACK sequence number */
        private HashMap<InetAddress, Integer> acks;

        /** map of InetAddress to last sent sequence number */
        private HashMap<InetAddress, Integer> sent;

        /** map of InetAddresses to acks to wait for */
        private HashMap<InetAddress, Integer> acksToWaitFor;

        /** set of InetAddresses to finished races this round */
        private HashSet<InetAddress> finishedRaces;

        /** scorecard of races -- if value is "false" then it
         * lost a race and should not be used */
        private HashMap<InetAddress, Boolean> scorecard;

        /** last time scorecard was cleared */
        private long lastScorecardUpdate = -1;

        /** start time for current race */
        private long raceStartTime = -1;

        /** last time race was performed */
        private long lastRaceTime = -1;

        /** bytes for destination for data in this experiment */
        byte[] dest;

        /** port for other end */
        int endpointPort;

        /** maximum number of packets in flight */
        private final int MAX_QUEUE = 4;

        /** if true, will force the primary path to be slower than the detour path */
        boolean forceDelay = false;

        /** allows for tight control of when such delays happen */
        int opportunitiesToDelay = 0;

        /** controls whether the stream can perform a race */
        private boolean canCheckRace = false;
        
        /** the timer for probing for nearby peers */
        ITimer edgeServerListTimer;

		private boolean done = false;

		private InetAddress currentInetAddress;

		private InetAddress directInetAddress;

		private long lastSendTime = -1;

		private ITimer timeoutTimer;
		
		private boolean transferFailed = false;
		
	    /** maps an experiment number of a set of pairs of values, the
	     * pairs being the number of bytes sent and the time they were sent
	     * for each message sent */
	    private HashMap<InetAddress, Vector<Pair<Long, Long>>> senderDataTransferStats;

		private long lastTPCompareTime;

		private OutputStream myOutputStream;

		private InputStream myInputStream;

		private Object dataTransferObject = new Object();

		private boolean pathChanged = false;

		private InetAddress newPath = null;
		

        /**
         *
         * @param onoExp the experiment for this stream
         * @param directConnection the direct connection between source and dest
         * @param parent the current data transfer object
         */
        public StreamHandler(DataTransferConfig onoExp,
        		IOnoConnection directConnection, OnoDataTransferManager parent) {
            this.onoExp = onoExp;
            forceDelay = onoExp.forceDelay;
            this.directConnection = directConnection;

            InetSocketAddress sock = directConnection.getEndpoint()
                                                     .getNotionalAddress();
            otherEnd = sock.getAddress().toString().substring(1);
            this.parent = parent;
            currentConnection = directConnection;
            currentInetAddress = getEndpointInetAddress(currentConnection);
            directInetAddress = currentInetAddress;
            myRaces = new HashMap<IOnoConnection, Long>();
            dest = null;
            lastTPCompareTime = System.currentTimeMillis();
            senderDataTransferStats =  new HashMap<InetAddress, Vector<Pair<Long, Long>>>();

            senderDataTransferStats.put(currentInetAddress, 
            		new Vector<Pair<Long, Long>>());

            try {
                dest = Address.getByAddress(onoExp.dest).getAddress();
            } catch (UnknownHostException e1) {
                // TODO Auto-generated catch block
                e1.printStackTrace();
            }

            endpointPort = sock.getPort(); // was a call to "getPort(onoExp)"
            acks = new HashMap<InetAddress, Integer>();
            sent = new HashMap<InetAddress, Integer>();
            acksToWaitFor = new HashMap<InetAddress, Integer>();
            finishedRaces = new HashSet<InetAddress>();
            scorecard = new HashMap<InetAddress, Boolean>();
            lastScorecardUpdate = System.currentTimeMillis();
            lastRaceTime = System.currentTimeMillis()-30*1000; // make it 30 seconds until first race
            
            edgeServerListTimer = MainGeneric.createTimer("EdgeServerListTimer");
//          UTTimer edgeServerListTimer = pi.getUtilities()
//                                          .createTimer("EdgeServerListTimer");
            EdgeServerListHandler handler = new EdgeServerListHandler(directConnection);
	          edgeServerListTimer.addEvent(System.currentTimeMillis() +
	              1000, handler);
	          edgeServerListTimer.addPeriodicEvent(1 * 30 * 1000, handler);
	          
	          // timer to handle stalled experiments
	          lastSendTime = System.currentTimeMillis();
	          timeoutTimer = MainGeneric.createTimer("ExperimentTimeout");
	          timeoutTimer.addPeriodicEvent(30*1000, new ITimerEventPerformer(){

				public void perform(ITimerEvent event) {
					if (System.currentTimeMillis()-lastSendTime > (TIMEOUT*1000)){
						cancelStream();
					}
					
				}});

            //            lastRaceTime = System.currentTimeMillis();
        }

        public void cancelStream() {
        	transferFailed = true;
			done  = true;
			
		}
        
        public void doneWithStream() {
			done  = true;
			
		}

		/**
         * Main functionality.
         */
        public void run() {
        	
        	long dataTransferred = 0;
            // repeat for each data transfer (currently not used and not tested)
            for (int j = 0; j < onoExp.numFileTransfers; j++) {
                // figure out how much data there is to send
                long dataLeftToSend = 0;
                int readVal=0;
                ByteBuffer outputBuffer = ByteBuffer.allocateDirect(65536*4);
            	outputBuffer.position(0);
            	int startPosition = 0;
                // loop until done...
                while (!done) {
                	
                	//startPosition = outputBuffer.position();
                    // clear scorecard if necessary
                    if ((System.currentTimeMillis() - lastScorecardUpdate) > SCORECARD_REFRESH) {
                        scorecard.clear();
                        lastScorecardUpdate = System.currentTimeMillis();
                    }

                    // make sure that no bad connections are stalling us here
                    cleanUpConnections();

                    int totalSentThisRound = 0;
                    int sentThisRound = 0;

                    // if detouring, check for nodes to connect to 
                    // and try to race if one is available
                    if (onoExp.useDetouring) {
                    	if (pathChanged ){
                    		setCurrentPath(newPath , currentInetAddress);
                    		pathChanged = false;
                    		newPath = null;
                    	}
                    	addNewMiddleNodes();
                    	
                    	
                        checkForMiddleNodesToConnect();

                        // update race paths
                        checkForRacesToConduct();
                    }

                    if (myInputStream==null){
                    	try {
							Thread.sleep(1000);
						} catch (InterruptedException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
                    	continue;
                    }
                    
                    try {
                    	
                    		int sleepCount = 0;
                    		int sizeToTry = OnoMessageData.payloadMaxSize(
                            		manager.getConnection(currentInetAddress));
                    	   if (sizeToTry == -1) {
                       			transferFailed=true;
                       			done = true;
                       			break;
                       		}
                    	   
                    	   sizeToTry = 4*(65536-9); // TODO fix hack here
                    	   
                    	   
                    	   	outputBuffer.limit(sizeToTry);
	                   		outputBuffer.position(0);
	                   		long totalRead = dataLeftToSend;
	                   		
	                   		do {
							
                    		while (totalRead < sizeToTry){
                    			readVal = myInputStream.read();
                    			if (readVal<0){
                    				if (totalRead==0){
                    					//done=true;
                    				}
                    				break;
                    			}
                    			outputBuffer.put((byte)readVal);
                    			totalRead++;
                    			dataTransferred++;
		                    }
                    	
		                    dataLeftToSend = totalRead;
		                    
//		                    if (dataLeftToSend==0 || myInputStream.available()==0){
//		                    	try {
//		                    		// TODO hacky
//									Thread.sleep(1);
//									sleepCount++;
//								} catch (InterruptedException e) {
//						
//									e.printStackTrace();
//								}
//		                    }
                    	}   while (readVal >= 0 && dataLeftToSend<sizeToTry);
	                   		outputBuffer.position(0);
	                   			
	                   		//System.out.println("Slept " + sleepCount + " times!");
                    } catch(IOException e){
                    	e.printStackTrace();
                    }
                    
                    
                    
                    int dataToSend;

                    // TODO send app data over detour path when racing, too
                    // send over detour paths
                    if (onoExp.useDetouring) { // only do this if detouring is enabled

                        synchronized (myRaces) {
                            // TODO what happens if all race data is sent immediately while the 
                            // other data is metered?
                            LinkedList<IOnoConnection> toRemove = new LinkedList<IOnoConnection>();

                            for (Entry<IOnoConnection, Long> e : myRaces.entrySet()) {
                                // TODO this may be called when it shouldn't...
                                // reset the start time
                                if ((raceStartTime == -1) &&
                                        (e.getValue() > 0)) {
                                    System.out.println(
                                        "Setting new start time!");
                                    raceStartTime = System.currentTimeMillis();
                                }

                                // TODO fix this stuff
                                InetAddress endpoint = getEndpointInetAddress(e.getKey());

                                if (acks.get(endpoint) == null) {
                                    acks.put(endpoint, 0);
                                }

                                if (sent.get(endpoint) == null) {
                                    sent.put(endpoint, 0);
                                }

                                if (sent.get(endpoint) > (acks.get(endpoint) +
                                		MAX_QUEUE)) {
                                    // TODO this hack sucks. it's because some races never finish 
                                    // for whatever reason
                                    acks.put(endpoint, sent.get(endpoint));
                                }

                                // figure out how much to send
                                sentThisRound = acks.get(endpoint) -
                                    sent.get(endpoint) + MAX_QUEUE;

                                // increase window if forced delay is enabled for direct path
                                if (shouldForceDelay()) {
                                    sentThisRound += (INTERVAL - MAX_QUEUE);
                                }

                                if (sentThisRound > 0) {
                                    if (shouldForceDelay() &&
                                            (acksToWaitFor.get(endpoint) == null)) {
                                        acksToWaitFor.put(endpoint,
                                            acks.get(endpoint) + INTERVAL);
                                    }

                                    if (sentThisRound > (MAX_QUEUE + INTERVAL)) {
                                        System.err.println(
                                            "Too many packets sent!");
                                    }
                                }

                                if (sentThisRound < 0) {
                                    sentThisRound = 0;
                                }

                                // start sending bytes
                                for (int k = 0;
                                        (k < sentThisRound) &&
                                        (e.getValue() >= 0); k++) {
                                    //		                            System.out.println("Found data to send...");
                                    // make sure that we should send data
                                    if (!isRaceExpired() &&
                                            ((dataLeftToSend > 0) &&
                                            (e.getValue() > 0))) {
                                        dataToSend = (int) Math.min(e.getValue(),
                                                (long)getRacePayloadSize()); // sending 1k byte packets for tighter control over warmup period
                                                                       // send just one byte at first to get basis for measurement

                                        if (e.getValue().intValue() == SIZE_EVAL) {
                                            dataToSend = 1;
                                            if (DO_TRACEROUTE){
                                            	TraceRouteRunner.getInstance().addIp(
                                            			endpoint.getHostAddress());
                                            	TraceRouteRunner.getInstance().addIp(                                            
                                                		currentInetAddress.getHostAddress());
                                            }
                                        }

                                        // TODO send app data when I can ensure that it makes it 
                                        // to the endpoint
                                        sendData(getEndpointInetAddress(e.getKey()), dataToSend,
                                            TYPE_RACE, dest, MainGeneric.getPeerPort(onoExp.dest), null);
                                        
                                        senderDataTransferStats.get(getEndpointInetAddress(e.getKey())).
                                        add(new Pair<Long, Long>((long)dataToSend, System.currentTimeMillis()));
                                        
                                        if (STREAM_DEBUG) System.out.println("(#" +
                                            (sent.get(endpoint) + k + 1) + ")");
                                        e.setValue(e.getValue() - dataToSend);
                                    } else { // we couldn't send data, maybe because the race period is over.
                                        e.setValue(0L);
                                    }

                                    // add to set for removal if done
                                    Integer comparison = compareThroughputWindows(SERIOUS_DROP, 3, false, endpoint);
                                    if (e.getValue() == 0 || (comparison!=null && comparison==0)) {
                                    	e.setValue(0L);
                                        if ((acks.get(endpoint) == null) ||
                                                (acks.get(endpoint) == 0)) {
                                            System.out.println(
                                                "No data sent during race!");

                                            break; // this didn't get to send any data in time
                                        } else { // end race
                                            sendDone(e.getKey(), TYPE_RACE);

                                            sentThisRound = k + 2;
                                            opportunitiesToDelay++;
                                        }

                                        toRemove.add(e.getKey());
                                        e.setValue(-1L);
                                        raceStartTime = -1;

                                        break;
                                    }
                                } // end for data to send

                                if (sent.get(endpoint) != null) {
                                    sent.put(endpoint,
                                        sent.get(endpoint) + sentThisRound);
                                }

                                synchronized(middleNodes){
                                // update time last used
                                middleNodes.get(onoExp.experimentId)
                                           .put(e.getKey().getEndpoint()
                                                 .getNotionalAddress(),
                                    System.currentTimeMillis());
                                }
                            } // end for each race

                            totalSentThisRound += sentThisRound; // update total sent this round

                            // add finished races to list of ones to remove
                            for (IOnoConnection con : toRemove) {
                                finishedRaces.add(getEndpointInetAddress(con));
                            }

                            // don't close the connection, otherwise data won't be sent
                        } // end monitor on myRaces
                    } // end sending detour data

                    // don't send over direct path until ack is at proper index...
                    if (acksHaveCleared()) {
                        // send direct
                        synchronized (dataTransferObject) {
                            InetAddress endpoint = currentInetAddress;

                            if (acks.get(endpoint) == null) {
                                acks.put(endpoint, 0);
                            }

                            if (sent.get(endpoint) == null) {
                                sent.put(endpoint, acks.get(endpoint));
                            }

                            synchronized (acks) {
                                sentThisRound = Math.min(acks.get(endpoint) -
                                    sent.get(endpoint) + MAX_QUEUE, 1);

                                if (sentThisRound > MAX_QUEUE) {
                                	/// TODO why is this happening?
                                    System.out.println(
                                        "Sending too many packets!");
                                    sent.put(endpoint,acks.get(endpoint));
                                }
                            }
                            
                            // send data on primary path
                            for (int i = 0; i < sentThisRound; i++) {
                                dataToSend = (int)Math.min(dataLeftToSend,
                                        (long)OnoMessageData.payloadMaxSize(
                                        		manager.getConnection(currentInetAddress)));

                                // just send one byte at first to get basis for speed
                                if (dataLeftToSend == onoExp.size) {
                                    dataToSend = 1;
                                    if (DO_TRACEROUTE){
                                    	TraceRouteRunner.getInstance().addIp(currentInetAddress.getHostAddress());
                                    }
                                }
                                
                                // update experiment progress
                                lastSendTime  = System.currentTimeMillis();

                                //outputBuffer.position(startPosition);
                                // if direct path, no need to specify dest
                                if (currentInetAddress.equals(directInetAddress)) {
                                    sendData(currentInetAddress, dataToSend,
                                        TYPE_PRIMARY_PATH, DIRECT_BYTES,
                                        endpointPort, outputBuffer);
                                } else { // otherwise, it's going to a middle hop, so specify the dest
                                    sendData(currentInetAddress, dataToSend,
                                        TYPE_PRIMARY_PATH, dest, MainGeneric.getPeerPort(onoExp.dest), 
                                        outputBuffer);
                                }

                                senderDataTransferStats.get(currentInetAddress).
                                add(new Pair<Long, Long>((long)dataToSend, System.currentTimeMillis()));
                                
                                if (sent.get(endpoint) != null) {
                                    if (STREAM_DEBUG) System.out.println("(#" +
                                        (sent.get(endpoint) + i + 1) + ")");
                                }

                                if (STREAM_DEBUG) {
                                    debugRaces();
                                }

                                // update data left to send
                                dataLeftToSend -= dataToSend;
                                if (STREAM_DEBUG) System.out.println("Data left to send: " +
                                    dataLeftToSend);

                                // if done, end transfer
                                if (readVal < 0) {
                                    handleDone(endpoint);

                                    break;
                                } 
                                if (dataLeftToSend==0) break;
                            }

                            if (sent.get(endpoint) != null) {
                                sent.put(endpoint,
                                    sent.get(endpoint) + sentThisRound);
                            }
                            
                            
                        }
                    }

                    totalSentThisRound += sentThisRound;

                    // if nothing has been sent, sleep for a short while and try again
                    if (totalSentThisRound <= 0) {
                        try {
                            Thread.sleep(100);
                        } catch (InterruptedException e) {
                            // TODO Auto-generated catch block
                            e.printStackTrace();
                        }
                    }
                } // end data transfer

                // record data
                if (!transferFailed) Statistics.getInstance().commitDataTransfers();
                    
            } // end loop over all experiments
            
            System.out.println("Done with data transfer (sent "+ dataTransferred + " bytes)...waiting!");
            // wait a little bit before doing this...
            try {
            	
                Thread.sleep(15 * 1000);
                //myOutputStream.flush();	
            } catch (InterruptedException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            } 
//            catch (IOException e) {
//				// TODO Auto-generated catch block
//				e.printStackTrace();
//			}

            System.out.println("Cleanup!");
            // cleanup!
            
            edgeServerListTimer.destroy();
            myDestinations.remove(otherEnd);
            currentPrimary.remove(onoExp.experimentId);
            synchronized(readyToRace){readyToRace.remove(onoExp.experimentId);}
            synchronized(middleNodes){middleNodes.remove(onoExp.experimentId);}
            //experiments.remove(0);
            streamHandlers.remove(onoExp.experimentId);
            try {
						
	            myInputStream.close();
	            //myOutputStream.close();
            } catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
            for (IOnoConnection con : manager.getActiveConnections().values()) {
                con.close();
            }
            
            synchronized(this){
            	notifyAll();
            }


            // TODO close connection?
        }

		private void handleDone(InetAddress endpoint) {
			System.out.println("Sending done...");
			sendDone(manager.getConnection(currentInetAddress),
			    TYPE_PRIMARY_PATH);

			try {
			    Thread.sleep(1 * 1000);
			} catch (InterruptedException e) {
			    // TODO Auto-generated catch block
			    e.printStackTrace();
			}

			done = true;

			synchronized (acks) {
			    acks.remove(endpoint);
			}

			sent.remove(endpoint);
		}

        private void addNewMiddleNodes() {
        	synchronized(middleNodes){
        	if (middleNodes.get(onoExp.experimentId) == null) {
                middleNodes.put(onoExp.experimentId,
                    new LinkedHashMap<InetSocketAddress, Long>());
            }
        	}

        	
        	
        	if (peersCloseToMe == null){
        		if (HACK_MIDDLE  ){
        			peersCloseToMe = new HashSet<InetSocketAddress>();
        			peersCloseToMe.add(new InetSocketAddress("24.12.67.232", 
        				MainGeneric.getPeerPort("24.12.67.232")));

                }
        		else return;
        	}
        	
            for (InetSocketAddress sock : peersCloseToMe ) {
            	synchronized(middleNodes){
                if (middleNodes.get(onoExp.experimentId)
                                   .get(sock) == null) {
                    middleNodes.get(onoExp.experimentId)
                               .put(sock, -1L);
                }
            	}
            }
			
		}

		/**
        * @return
        */
        private boolean isRaceExpired() {
            boolean raceExpired = false;

            if (raceStartTime > 0) {
                raceExpired = (System.currentTimeMillis() - raceStartTime) > (WARMUP_TIME +
                    TEST_TIME);
            }

            //            if (!raceExpired) System.out.println("Race not expired!");
            //            else{
            //                /*System.out.println("Race expired!");
            //                System.out.println("Race start: "+raceStartTime + "\tCurrent time:"+
            //                        System.currentTimeMillis());*/
            //            }
            return raceExpired;
        }

        /**
         * @return
         */
        private int getRacePayloadSize() {
            //          OnoDataMessage.payloadMaxSize(e.getKey())
            int size = 0;

            if ((System.currentTimeMillis() - raceStartTime) < WARMUP_TIME) {
                size = WARMUP_SIZE;
            } else {
                size = NON_WARMPUP_SIZE;
            }

            return size;
        }

        /**
         * @return
         */
        private boolean shouldForceDelay() {
            return forceDelay && (opportunitiesToDelay == 1);
        }
        
        public boolean isDone(){
        	return done;
        }

        private void cleanUpConnections() {
            HashSet<InetAddress> ackAddrs = new HashSet<InetAddress>();

            for (InetAddress addr : acksToWaitFor.keySet()) {
                ackAddrs.add(addr);
            }

            for (InetSocketAddress sock : manager.getActiveConnections().keySet()) {
                if (ackAddrs.contains(sock.getAddress())) {
                    ackAddrs.remove(sock.getAddress());
                }
            }

            for (InetAddress addr : ackAddrs) {
                sent.remove(addr);
                acks.remove(addr);
                acksToWaitFor.remove(addr);
            }

            if (isRaceExpired()) {
                if (myRaces.size() > 0) {
                    Entry<IOnoConnection, Long> e = myRaces.entrySet()
                                                                        .iterator()
                                                                        .next();
                    InetAddress end = getEndpointInetAddress(e.getKey());

                    if ((e.getValue() == SIZE_EVAL) ||
                            ((acks.get(end) == 0) && (raceStartTime != -1))) {
                        synchronized(myRaces){myRaces.clear();}
                        acks.remove(end);
                        sent.remove(end);
                        acksToWaitFor.remove(end);
                        raceStartTime = -1;
                        scorecard.put(end, false);
                    }
                }
            }
            
            // TODO remove when errant code is gone
//            if (manager.getConnection(currentInetAddress)!=currentConnection){
//            	System.err.println("Weird connection reference behavior!");
//            	if (directConnection == currentConnection){
//            		directConnection = manager.getConnection(currentConnection.getEndpoint().getNotionalAddress().getAddress());
//            	}
//            	currentConnection = manager.getConnection(currentConnection.getEndpoint().getNotionalAddress().getAddress());
//            }
        }

        private boolean acksHaveCleared() {
        	if (!MainGeneric.isAzureus()) return true;
            boolean ret = true;

            synchronized (acks) {
                for (Entry<InetAddress, Integer> ent : acksToWaitFor.entrySet()) {
                    Integer ackVal = acks.get(ent.getKey());
                    Integer ackToWaitFor = ent.getValue();

                    if ((acks.get(ent.getKey()) == null) &&
                            (sent.get(ent.getKey()) != null)) {
                        return false;
                    } else if ((acks.get(ent.getKey()) != null) &&
                            (ackToWaitFor > ackVal)) {
                        return false;
                    }
                } // end iteration

                // if we get here, then all acks have cleared
                acksToWaitFor.clear();
            }

            return ret;
        }

        private void sendDone(IOnoConnection con, byte type) {
            int size = 9;
            //if (type == TYPE_RACE){
            size += 8;

            //}
            
            ByteBuffer bb = manager.getByteBuffer(size);
            bb.put(type);
            //if (type==TYPE_RACE){
            bb.put(dest);
            bb.putInt(MainGeneric.getPeerPort(onoExp.dest));

            //}
            String s = "done";

            for (int i = 0; i < s.length(); i++)
                bb.putChar(s.charAt(i));

            bb.position(0);

            try {
                con.send(bb);
            } catch (IMessageException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }

        private void checkForRacesToConduct() {
            // first check if on and not yet time to do another race
        	if (sent.size()>=MAX_CONCURRENT_RACES) return;
        	if (dropInThroughput()){
        		// don't return
        	}
        	else if (!onoExp.randomPeers &&
                    (!canCheckRace ||
                    ((System.currentTimeMillis() - lastRaceTime) < ANY_PROBE_INTERVAL))) {
                return;
            }

            // if not using direct path, first check to see 
            // if we are doing random peers
            if (onoExp.randomPeers) {
                if (myRaces.size() > 0) {
                    return;
                }

                synchronized(readyToRace){
	                LinkedList<IOnoConnection> peers = readyToRace.get(onoExp.experimentId);
	
	                if ((peers == null) || (peers.size() == 0)) {
	                    return;
	                }
	
                    synchronized(myRaces){myRaces.put(peers.get(r.nextInt(peers.size())), SIZE_EVAL);}
                }
                lastRaceTime = System.currentTimeMillis();

                return;
            }

            // handle periodic checking of direct path
            if ((currentInetAddress != directInetAddress) &&
                    (myRaces.size() == 0)) {
                InetSocketAddress sock = manager.getConnection(directInetAddress).getEndpoint()
                                                         .getNotionalAddress();

                synchronized(middleNodes){
                	if (middleNodes.get(onoExp.experimentId).get(sock) == null) {
                		middleNodes.get(onoExp.experimentId)
                		.put(sock, System.currentTimeMillis());
                	}


                	if ((System.currentTimeMillis() -
                			middleNodes.get(onoExp.experimentId).get(sock)) > PEER_PROBE_INTERVAL) {
                		middleNodes.get(onoExp.experimentId)
                		.put(sock, System.currentTimeMillis());
                        synchronized(myRaces){
                            myRaces.put(manager.getConnection(directInetAddress), SIZE_EVAL);
                        }
                		lastRaceTime = System.currentTimeMillis();

                		return;
                	}
                }
            }
            // get rid of nodes we don't care about
            LinkedList<IOnoConnection> validForRaces = new LinkedList<IOnoConnection>();
            LinkedList<IOnoConnection> middles;

            synchronized (readyToRace) {
                middles = readyToRace.get(onoExp.experimentId);

                // first remove all that i do not want to deal with
                if (middles == null) {
                    return;
                }

                for (IOnoConnection con : middles) {
                    if (scorecard.containsKey(getEndpointInetAddress(con))) {
                        continue;
                    }
                    // don't allow the current connection either.
                    else if (getEndpointInetAddress(con)
                                     .equals(currentInetAddress)) {
                        continue;
                    }
                    else if (sameClassCSubnet(getEndpointInetAddress(con), currentInetAddress)){
                    	continue;
                    }
                    validForRaces.add(con);
                }

                if (validForRaces.size() == 0) {
                    return; // no middles to pick from
                }

                synchronized (myRaces) {
                	IOnoConnection newCon = null;

                    if (myRaces.size() > 0) {
                        return; // don't want to add if currently going
                    }

                    LinkedHashSet<InetAddress> potentialOnes = new LinkedHashSet<InetAddress>();

//                    for (HashSet<InetSocketAddress> set : peersPerEdge.values()) {
//                        for (InetSocketAddress sock : set) {
//                            potentialOnes.add(sock.getAddress());
//                        }
//                    }
//
//                    if (potentialOnes.size() == 0) {
//                        return;
//                    }

                    LinkedList<IOnoConnection> middlesToTry = new LinkedList<IOnoConnection>();

                    for (IOnoConnection con : validForRaces) {
//                        if (potentialOnes.contains(con.getEndpoint()
//                                                          .getNotionalAddress()
//                                                          .getAddress())) {
                            middlesToTry.add(con);
//                        }
                    }

                    synchronized(middleNodes){
                    if (middlesToTry.size() == 0) {
                        for (IOnoConnection con : middles) {
                            InetSocketAddress sock = con.getEndpoint()
                                                        .getNotionalAddress();
                            middleNodes.get(onoExp.experimentId).put(sock, -1L);
                        }

                        return;
                    }
                    }

                    // now pick a random new guy
                    if (newCon == null) {
                        newCon = middlesToTry.get(r.nextInt(middlesToTry.size()));
                    }
                    
                    if (HACK_MIDDLE){
                    	try {
							if (manager.getConnection(InetAddress.getByName("24.12.67.232"))!=null){
								newCon = manager.getConnection(InetAddress.getByName("24.12.67.232"));
							}
						} catch (UnknownHostException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
                    }
                    senderDataTransferStats.put(getEndpointInetAddress(newCon), 
                    		new Vector<Pair<Long, Long>>());
                    myRaces.put(newCon, SIZE_EVAL);
                    lastRaceTime = System.currentTimeMillis();
                }
            }

            canCheckRace = false; // turn it back off
        }

        /**
         * Returns true if throughput has dropped 
         * by the maximum allowable amount over the past
         * window.
         * @return
         */
        private boolean dropInThroughput() {
        	
			if (lastTPCompareTime > System.currentTimeMillis()-WINDOW_SIZE*1000) return false;
			else lastTPCompareTime = System.currentTimeMillis();
			Integer returned = compareThroughputWindows(SERIOUS_DROP, WINDOW_SIZE, true, currentInetAddress);
			if (returned == null) return false;
			else return returned<0;
			
			
		}
    
        /**
         * Compares throughput in two windows: one from now until windowSize
         * seconds ago and one from windowSize seconds ago until 2*windowSize 
         * units ago.
         * 
         * Returns -1 if the the throughput has dropped below the specified 
         * threshold.
         * Returns 0 if the throughput has not changed beyond the threshold
         * Returns 1 if the throughput has increased beyond the threshold
         * Returns null if the comparison can't be performed.
         * 
         * @param threshold
         * @param windowSize
         * @param deleteData if true, deletes data outside the specified window
         * @return
         */
    	private Integer compareThroughputWindows(double threshold, int windowSize, 
    			boolean deleteData, InetAddress address){
    		int totalWindow = windowSize*2;
        	long currentTime = System.currentTimeMillis();
        	long minTime = currentTime - totalWindow*1000;
        	long middleTime = currentTime-windowSize*1000;
			synchronized (dataTransferObject){
				// not enough data
				Vector<Pair<Long, Long>> data = 
					senderDataTransferStats.get(address);
				if (data.size()<3) return null;
				if (data.get(0).getValue() > minTime)
					return null;
				
				// now there is at least enough data to look at...
				int startIndex = -1;
				Pair<Long, Long> p;
				for (int i = 0; i < data.size(); i++){
					p = data.get(i);
					// data is too small or too old
					if (p.getKey()<= WARMUP_SIZE || p.getValue() < minTime ){
						continue;
					} else {
						// we can make a measurement
						if (data.size() - (i-1) >= 3){
							startIndex = i-1;
							// clear out old data
							if (i>1 && deleteData)
								for (int j = 0; j < i-1; j++)
									data.remove(0);
							break;
						} else {
							// clear out old data
							if (i>1 && deleteData)
								for (int j = 0; j < i-1; j++)
									data.remove(0);
							return null;
						}
					}
					
				}
				
				if (startIndex==-1) return null;
				
				// now we know the startindex, so work backwards.
				int currentIndex = data.size()-1;
				long startTime = data.get(currentIndex).getValue();
				long endTime = startTime;
				int dataSum = 0;
				
				for (int i = currentIndex-1; i > 0; i--){
					dataSum+=data.get(i).getKey();
					currentIndex = i;
					if (data.get(i).getValue() < middleTime){
						endTime = data.get(i).getValue();
						break;
					} 
					
				}
				if (startTime==endTime) return null;
				double oldRate = 1000*((double)dataSum)/(startTime-endTime);
				startTime = endTime;
				dataSum = 0;
				for (int i = currentIndex-1; i >= 0; i--){
					dataSum+=data.get(i).getKey();
				}
				endTime = data.get(0).getValue();
				if (startTime==endTime) return null;
				double newRate = 1000*((double)dataSum)/(startTime-endTime);
				
				if (newRate/oldRate < threshold){
					return -1;
				} else if (oldRate/newRate < threshold){
					return 1;
				} else 
					return 0;
			}
    	}

		/**
        * @return
        */
        private void makeAllOnoConnections() {
            if (listenerActivated) {
                return;
            }

            //            for (Download d : pi.getDownloadManager().getDownloads()){
            //                if (d.getName().contains("OnoExp")){
            //                    
            //                    if (d.getPeerManager()!=null){
            //                        
            //                        d.addTrackerListener(new DownloadTrackerListener(){
            //
            //                            public void scrapeResult(DownloadScrapeResult result) {
            //                                // TODO Auto-generated method stub
            //                                
            //                            }
            //
            //                            public void announceResult(DownloadAnnounceResult result) {
            //                                // TODO Auto-generated method stub
            //                                
            //                                for (DownloadAnnounceResultPeer p : result.getPeers()){
            //                                    
            //                                    String peerIp = p.getAddress();                        
            //                                    int port = p.getPort(); 
            //                                    initiateConnection(new InetSocketAddress(peerIp, port));                                    
            //                            }}});
            //                        d.requestTrackerAnnounce(true);
            //                        listenerActivated = true;
            //                 
            //                    }
            //                }
            //            }
            if (otherPeers.size() > 0) {
                for (InetAddress addr : otherPeers) {
                    initiateConnection(new InetSocketAddress(addr, MainGeneric.getPeerPort(onoExp.dest)));
                }

                listenerActivated = true;
            }
        }

        private void checkForMiddleNodesToConnect() {
            // TODO periodically check direct path if currently using alternate
            synchronized (middleNodes) {
                LinkedHashMap<InetSocketAddress, Long> peers = middleNodes.get(onoExp.experimentId);

                if (peers == null) {
                    return;
                }

                long currentTime = System.currentTimeMillis();

                if (manager.getConnection(directInetAddress)==null){
                    manager.connect(new InetSocketAddress(directInetAddress, MainGeneric.
                            getPeerPort(directInetAddress.getHostAddress())), parent, true);
                    return;
                }
                for (Entry<InetSocketAddress, Long> e : peers.entrySet()) {
                    if (e.getKey().getAddress()
                             .equals(manager.getConnection(directInetAddress).getEndpoint()
                                                         .getNotionalAddress()
                                                         .getAddress())) {
                        continue;
                    }

                    if ((currentTime - e.getValue()) > PEER_PROBE_INTERVAL) {
                        e.setValue(System.currentTimeMillis());
                        
                        IOnoConnection con = manager.getConnection(currentInetAddress);
                        if (con!=null && con.getEndpoint()!=null && con.getEndpoint().getNotionalAddress()!=null && 
                                !e.getKey()
                                  .equals(manager.getConnection(currentInetAddress).getEndpoint()
                                                               .getNotionalAddress())) {
                            initiateConnection(e.getKey());
                        }
                    }
                }
            }

            if (onoExp.randomPeers) {
                makeAllOnoConnections();
            }
        }

        private void initiateConnection(InetSocketAddress sock) {
//            try {
        	// TODO use delay if connection recently failed...
            	if (!manager.hasConnection(sock)){
            	manager.connect(sock, parent, true);
            	
//                for (IOnoConnection con : connections.values()) {
//                    if (con.getEndpoint().getNotionalAddress().getAddress()
//                               .equals(sock.getAddress())) {
//                        return;
//
//                        //                        synchronized(readyToRace){
//                        //		            		if (readyToRace.get(onoExp.experimentId)==null) 
//                        //		            			readyToRace.put(onoExp.experimentId, new LinkedList<GenericMessageConnection>());
//                        //                            System.out.println("Added " + sock + " to current races!");
//                        //                            addToCurrentRaces(onoExp.experimentId, con);
//                        //		            		return;
//                        //	            		}
//                    }
//                }
//
//                System.out.println(
//                    "Actually trying to establish connection to " + sock);
//
//                GenericMessageEndpoint gme = gmr.createEndpoint(sock);
//                gme.addTCP(sock);
//
//                // wait before making connection
//                //Thread.sleep(60*1000);
//                GenericMessageConnection con;
//
//                con = gmr.createConnection(gme);
//
//                con.addListener(parent);
//                con.connect();

                pendingMiddleConnections.put(sock.getAddress(),
                    onoExp.experimentId);
            	}
//                connecting.add(sock);
//            } catch (MessageException e) {
//                // TODO Auto-generated catch block
//                e.printStackTrace();
//            }
        }
        
        public void setCurrentPath(InetAddress winner, InetAddress loser) {
        	 InetAddress currentEndpoint = currentInetAddress;
             InetAddress directEndpoint = directInetAddress;
             // if the direct path won or the current endpoint lost
             if (winner.equals(currentEndpoint)) { // current path won

                 if (myRaces.size() > 0) {
                     

                     if (!loser.equals(directEndpoint)) { // current path was direct

                         if (!onoExp.randomPeers) {
                             scorecard.put(loser, false);
                         }
                     }
                 }
             } else if (!currentEndpoint.equals(directEndpoint)) { // current path lost
                

            	// the direct path won, so mark this middle path as bad
                if (!onoExp.randomPeers) {
                    scorecard.put(currentEndpoint, false);
                }

                System.out.println("Marking " + currentEndpoint +
                    " as bad link!");
            }
             
             if (!winner.equals(currentEndpoint)){
            	 sent.remove(currentEndpoint);
                 acks.remove(currentEndpoint);
             }
            
            for (Entry<InetSocketAddress, IOnoConnection> e : manager.getActiveConnections().entrySet()) {
                if (e.getKey().getAddress().equals(winner)) {
                    // if not direct path 
                    if (!winner.equals(directInetAddress)) {
                        // add direct path as "middle node" to try at next 
                        // interval 
                        synchronized (middleNodes) {
                            if (middleNodes.get(onoExp.experimentId)
                                               .get(manager.getConnection(directInetAddress).getEndpoint()
                                                                        .getNotionalAddress()) == null) {
                                // only add if not there already
                                LinkedHashMap<InetSocketAddress, Long> values;
                                values = middleNodes.remove(onoExp.experimentId);
                                middleNodes.put(onoExp.experimentId,
                                    new LinkedHashMap<InetSocketAddress, Long>());
                                middleNodes.get(onoExp.experimentId)
                                           .put(manager.getConnection(directInetAddress).getEndpoint()
                                                                .getNotionalAddress(),
                                    System.currentTimeMillis());
                                middleNodes.get(onoExp.experimentId)
                                           .putAll(values);
                            } else {
                                System.out.println(
                                    "Direct connection already added!");
                                System.out.println("Middle nodes: " +
                                    middleNodes);
                            }
                        }
                    }

                    if (acksToWaitFor != null) {
                        acksToWaitFor.remove(winner);
                        sent.remove(winner);
                        acks.remove(winner);

                        if (winner!=currentInetAddress){
	                        sent.remove(currentInetAddress);
	                        acks.remove(currentInetAddress);
	                        acksToWaitFor.remove(currentInetAddress);
                        }

                        for (InetAddress addr : finishedRaces) {
                            sent.remove(addr);
                            acks.remove(addr);
                            acksToWaitFor.remove(addr);
                        }

                        
                    }

                    // if connection has not changed, do not reset info
                    if (currentInetAddress.equals(winner)) {
                        if (STREAM_DEBUG) {
                            debugRaces();
                        }

                        return;
                    }

                    //sent.remove(getEndpointInetAddress(currentConnection));
                    //acks.remove(getEndpointInetAddress(currentConnection));

                    // even out the acks
                    
                    	senderDataTransferStats.get(currentInetAddress).clear();
                    	currentInetAddress = winner;
                    	System.err.println("Changed current address to "+currentInetAddress);

                        int amountToAdd = 1;

                        if (shouldForceDelay()) {
                            amountToAdd += INTERVAL;
                        }

                        //						sent.put(getEndpointInetAddress(currentConnection), 
                        //								acks.get(getEndpointInetAddress(currentConnection))+1);
                    

                    break;
                }
            
            }
            finishedRaces.clear();
            synchronized(myRaces){ myRaces.clear(); }
        }

        public void setCurrentPathExt(InetAddress winner, InetAddress loser) {
            InetAddress currentEndpoint = currentInetAddress;
            InetAddress directEndpoint = directInetAddress;
            
            
            	newPath = winner;
            	pathChanged = true;            	
            
           
            if (STREAM_DEBUG) {
                debugRaces();
            }
        }

        /**
         *
         */
        private void debugRaces() {
        	synchronized(middleNodes){
	            System.out.println("Middle nodes:\n" +
	                middleNodes.get(onoExp.experimentId));
        	}
            System.out.println("Ongoing races:");

            synchronized (myRaces){
                for (Entry<IOnoConnection, Long> ent : myRaces.entrySet()) {
                    System.out.println(ent.getKey().getEndpoint()
                                          .getNotionalAddress() + "\t" +
                        ent.getValue());
                }
            }

            if ((readyToRace != null) &&
                    (readyToRace.get(onoExp.experimentId) != null)) {
                System.out.println("Waiting races:");

                synchronized(readyToRace){
	                for (IOnoConnection con : readyToRace.get(
	                        onoExp.experimentId)) {
	                    System.out.print(con.getEndpoint().getNotionalAddress() +
	                        "\t");
	                }
                }

                System.out.print("\n");
            }

            System.out.println("Acks: " + acks);
            System.out.println("Sent: " + sent);
            System.out.println("Scorecard: " + scorecard);
            System.out.println("Mapped peers: " + peersPerEdge);
        }

        public void recordAck(InetAddress ina, int ackNum) {
            // TODO validate sequence numbers
//        	if (ackNum>=9999){
//        		System.out.println("Got ack number "+ackNum+"!");
//        	}
        	if (acks.get(ina)==null){
        		acks.put(ina, 0);
        		sent.put(ina, 0);
        	}
            if (acks.get(ina) != null) {
                synchronized (acks) {
                    if ((sent.get(ina) == null) ||
                            ((sent.get(ina) + MAX_QUEUE) < ackNum)) {
                    	String errString = "Invalid ack number for " + ina +
                        "!" + "\nSent: " + sent.get(ina) + "\tAck num: " +
                        ackNum;
                        System.err.println(errString);
                        reportError(errString);
                        // TODO solve timing issues that mess this up
                        sent.put(ina, sent.get(ina) - acks.get(ina) - 1 +
                            ackNum);
                    }

                    acks.put(ina, ackNum);
                }
            }
        }

        public void resetAcks(InetAddress endpointInetAddress) {
            synchronized (acks) {
                acks.remove(endpointInetAddress);
            }

            sent.remove(endpointInetAddress);
            acksToWaitFor.remove(endpointInetAddress);
        }

        public void removeRace(IOnoConnection con) {
            synchronized(myRaces){
                if (myRaces.remove(con) != null) {
                    
                    raceStartTime = -1;
                }
            }
        }

        /**
         *
         */
        public void allowRaceChecking() {
            canCheckRace = true;
        }

        /**
         * @param address
         * @return
         */
        public boolean peerCanBeAdded(InetAddress address) {
            // TODO Auto-generated method stub
            return !scorecard.containsKey(address);
        }

		public InetAddress getCurrentInetAddress() {
			return currentInetAddress;
		}

		public void setCurrentInetAddress(InetAddress currentInetAddress) {
			this.currentInetAddress = currentInetAddress;
		}

		public OutputStream getOutputStream() {
			CircularByteBuffer cbb = new CircularByteBuffer(
					1024*1024);
			
			
			if (myInputStream==null){
				
				
				myOutputStream = cbb.getOutputStream();
				myInputStream = cbb.getInputStream();
//				try {
//					myInputStream.connect(myOutputStream);
//				} catch (IOException e) {
//					return null;
//				}
				
			}
			return myOutputStream; 
		}

    }
    
    /**
     * @param connection
     * @return
     */
    private InetAddress getEndpointInetAddress(
        IOnoConnection connection) {
        return connection.getEndpoint().getNotionalAddress().getAddress();
    }
    
    public void reportError(String errString) {
		Statistics.getInstance().reportError(new Exception(errString));
		
	}

	public boolean sameClassCSubnet(InetAddress ip1, InetAddress ip2) {
		byte ip1b[] = ip1.getAddress();
		byte ip2b[] = ip2.getAddress();
		return ip1b[3]==ip2b[3] && ip1b[1]==ip2b[1] && ip1b[2]==ip2b[2];
	}

	/**
     * @param connection
     * @return
     */
    private String getEndpointIpAddress(IOnoConnection connection) {
        String endpoint = getEndpointInetAddress(connection).toString();

        return endpoint.substring(endpoint.indexOf('/') + 1);
    }

	public boolean isRunning() {
		return myDestinations.size()!=0 || (sources.size()>0 && 
                !sources.values().iterator().next().hasExpired());
	}

	public void cancelTimer() {
		if (timer!=null) timer.destroy();
		
	}




	public void closeConnections(Object transferContext) {
		
		DataTransferConfig config = transferToConfig.remove(transferContext);
		InetSocketAddress socket = transfers.remove(transferContext);		
		configToTransfer.remove(config);
		//experiments.remove(config);
//		if (socket!=null)
//			myDestinations.remove(socket.getAddress().getHostAddress());
//		if (config!=null){
//			if (streamHandlers.get(config.experimentId)!=null)
//				streamHandlers.get(config.experimentId).doneWithStream();
//		}
		
	}




	public InputStream getInputStream(Object transferContext) {
		synchronized(transferContext){
			while (!manager.hasConnection(transfers.get(transferContext))){
				try {
					transferContext.wait();
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}
		if (destInputStreams.get(transferContext)==null){
			CircularByteBuffer cbb = new CircularByteBuffer(
					CircularByteBuffer.INFINITE_SIZE);
//			PipedInputStream pis = new PipedInputStream();
//			PipedOutputStream pos = new PipedOutputStream();
//			try {
//				pis.connect(pos);
//			} catch (IOException e) {
//				// TODO Auto-generated catch block
//				return null;
//			}
			destInputStreams.put(transferContext, cbb.getInputStream());
			destOutputStreams.put(transferContext, cbb.getOutputStream());
		}
		return destInputStreams.get(transferContext);
	}




	private InetSocketAddress getSocketForTransferContext(Object transferContext) {
		transfers.get(transferContext);
		return null;
	}




	public OutputStream getOutputStream(Object transferContext) {
		// TODO hacky w/ sleep
		DataTransferConfig dtc = transferToConfig.get(transferContext);
		while (streamHandlers.get(dtc.experimentId)==null){
			try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		OutputStream toReturn = streamHandlers.get(dtc.experimentId).
		getOutputStream();
		
        MainGeneric
        .createThread("StreamHandler" + dtc.experimentId, streamHandlers.get(dtc.experimentId));

		return toReturn;
		

	}

	public void createIncomingDataConnection(InetSocketAddress socket, 
			Object transferContext) {



	}


	public void createDataConnection(InetSocketAddress socket, Object transferContext) {
		
	
		
		transfers.put(transferContext, socket);
		DataTransferConfig config = new DataTransferConfig();
		config.dest = MainGeneric.getPublicIpAddress();
		config.source = socket.getAddress().getHostAddress();
		config.experimentId = transferContext.hashCode();
		configToTransfer.put(config, transferContext);
		transferToConfig.put(transferContext, config);
		experiments.add(config);
	
		
		sources.put(socket.getAddress().getHostAddress(), config);
		
        ddb = MainGeneric.getDistributedDatabase();
        
        CDNClusterFinder.getInstance().setPreferredName(config.preferredName);
		
       

        myIp = Digger.getInstance().getIp();


        int port = getPort(config);

        //              if (port==-1){
        //              // try the default port for PL nodes
        //              port = 7000;
        //              }

        // start experiment by connecting to dest
        InetSocketAddress otherEnd = new InetSocketAddress(config.source,
                port);

        
        manager.connect(otherEnd, this, true);
	}




	public boolean accept(IOnoConnection connection) throws IMessageException {
		System.out.println("Incoming connection from " +
	            connection.getEndpoint().getNotionalAddress() + "!");

	        connection.addListener(this);
	        
	        System.out.println("Connected! ");

            DataTransferConfig dtc = myDestinations.get(getEndpointIpAddress(connection));

            // connection from middle hop
            if ((dtc == null)) {
               return true;               
            }


            StreamHandler sh = new StreamHandler(dtc, connection, this);
            streamHandlers.put(dtc.experimentId, sh);

            CDNClusterFinder.getInstance().addPeer(
            		MainGeneric.getClusterFinderObject(null), MainGeneric.getPublicIpAddress(), 7000);

	        return true;
	}




	public void listenForConnection(InetSocketAddress socket, Object transferContext) {
		MainGeneric.initialize();
		
		transfers.put(transferContext, socket);
		DataTransferConfig config = new DataTransferConfig();
		config.dest = socket.getAddress().getHostAddress();
		config.source = MainGeneric.getPublicIpAddress();
		config.experimentId = transferContext.hashCode();
		config.useDetouring = true;
		configToTransfer.put(config, transferContext);
		transferToConfig.put(transferContext, config);
		experiments.add(config);
		try {
			myInetAddress = InetAddress.getByName(config.source);
		} catch (UnknownHostException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		myDestinations.put(socket.getAddress().getHostAddress(), config);
		
		ddb = MainGeneric.getDistributedDatabase();
	        
	    CDNClusterFinder.getInstance().setPreferredName(config.preferredName);
	    
	    
        ITimer edgeServerListTimer = MainGeneric.createTimer("EdgeServerListTimer");
//      UTTimer edgeServerListTimer = pi.getUtilities()
//                                      .createTimer("EdgeServerListTimer");
      EdgeServerListHandler handler = new EdgeServerListHandler();
      edgeServerListTimer.addEvent(System.currentTimeMillis() +
          1000, handler);
      edgeServerListTimer.addPeriodicEvent(1 * 30 * 1000, handler);

		
	}




	public void waitForDone(Object transferContext) {
		if (transferToConfig.get(transferContext)==null) 
			return;
		if ( streamHandlers.get(transferToConfig.get(transferContext).experimentId)==null) 
			return;
		StreamHandler sh = streamHandlers.get(transferToConfig.get(transferContext).experimentId);
		synchronized (sh){
			try {
				sh.wait();
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
	}
	
    
}
